"""
Show materialized configuration of tox environments.
"""

from textwrap import indent
from typing import Iterable, List

from tox.config.cli.parser import ToxParser
from tox.config.loader.stringify import stringify
from tox.config.sets import ConfigSet
from tox.plugin.impl import impl
from tox.session.common import env_list_flag
from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("config", ["c"], "show tox configuration", show_config)
    our.add_argument("-d", action="store_true", help="list just default envs", dest="list_default_only")
    our.add_argument(
        "-k", nargs="+", help="list just configuration keys specified", dest="list_keys_only", default=[], metavar="key"
    )
    our.add_argument(
        "--core", action="store_true", help="show core options too when selecting an env with -e", dest="show_core"
    )
    env_list_flag(our)


def show_config(state: State) -> int:
    keys: List[str] = state.options.list_keys_only
    # environments may define core configuration flags, so we must exhaust first the environments to tell the core part
    for name in state.env_list(everything=False):
        tox_env = state.tox_env(name)
        print(f"[testenv:{name}]")
        if not keys:
            print(f"type = {type(tox_env).__name__}")
        print_conf(tox_env.conf, keys)
        print("")
    # no print core
    if state.options.env.all or state.options.show_core:
        print("[tox]")
        print_conf(state.conf.core, keys)
    return 0


def print_conf(conf: ConfigSet, keys: Iterable[str]) -> None:
    for key in keys if keys else conf:
        if key not in conf:
            continue
        try:
            value = conf[key]
        except Exception as exception:  # because e.g. the interpreter cannot be found
            as_str, multi_line = f"# Exception: {exception!r}", False
        else:
            as_str, multi_line = stringify(value)
        if multi_line and "\n" not in as_str:
            multi_line = False
        if multi_line and as_str.strip():
            print(f"{key} =\n{indent(as_str, prefix='  ')}")
        else:
            print(f"{key} ={' ' if as_str else ''}{as_str}")
    unused = conf.unused()
    if unused and not keys:
        print(f"# !!! unused: {', '.join(unused)}")



from enum import Enum
from pathlib import Path
from typing import Any, Mapping, Sequence, Set, Tuple

from tox.config.types import Command, EnvList


def stringify(value: Any) -> Tuple[str, bool]:
    """
    Transform a value into a string representation.

    :param value: the value in question
    :return: a tuple, first the value as str, second a flag if the value if a multi-line one
    """
    if isinstance(value, str):
        return value, False
    if isinstance(value, (Path, float, int, bool)):
        return str(value), False
    if isinstance(value, Mapping):
        return "\n".join(f"{stringify(k)[0]}={stringify(v)[0]}" for k, v in value.items()), True
    if isinstance(value, (Sequence, Set)):
        return "\n".join(stringify(i)[0] for i in value), True
    if isinstance(value, Enum):
        return value.name, False
    if isinstance(value, EnvList):
        return "\n".join(e for e in value.envs), True
    if isinstance(value, Command):
        return value.shell, True
    return str(value), False


__all__ = ("stringify",)



"""
Display the version information about tox.
"""
from pathlib import Path

from tox.config.cli.parser import ToxParser
from tox.plugin.impl import impl


@impl
def tox_add_option(parser: ToxParser) -> None:
    import tox
    from tox.version import __version__

    parser.add_argument(
        "--version",
        action="version",
        version=f"{__version__} from {Path(tox.__file__).absolute()}",
    )



import re
from pathlib import Path

import pytest

from tox.pytest import MonkeyPatch
from tox.tox_env.python.req_file import RequirementsFile


@pytest.mark.parametrize(
    ["req", "key"],
    [
        ("--pre", "--pre"),
        ("--no-index", "--no-index"),
        ("--prefer-binary", "--prefer-binary"),
        ("--require-hashes", "--require-hashes"),
        ("  --pre", "--pre"),
        ("--pre ", "--pre"),
        (" --pre ", "--pre"),
        ("--pre\\\n", "--pre"),
        ("--pre # magic", "--pre"),
        ("--pre\t# magic", "--pre"),
        ("--no-index", "--no-index"),
        ("--find-links /my/local/archives", "--find-links /my/local/archives"),
        ("--find-links \\\n/my/local/archives", "--find-links /my/local/archives"),
        ("--find-links http://some.archives.com/archives", "--find-links http://some.archives.com/archives"),
        ("-i a", "-i a"),
        ("--index-url a", "--index-url a"),
        ("--extra-index-url a", "--extra-index-url a"),
        ("-e a", "-e a"),
        ("--editable a", "--editable a"),
        ("-c a", "-c a"),
        ("--constraint a", "--constraint a"),
        ("-f a", "-f a"),
        ("--find-links a", "--find-links a"),
        ("--trusted-host a", "--trusted-host a"),
        ("--use-feature a", "--use-feature a"),
        ("--no-binary a", "--no-binary a"),
        ("--only-binary a", "--only-binary a"),
        ("####### example-requirements.txt #######", ""),
        ("\t##### Requirements without Version Specifiers ######", ""),
        ("  # start", ""),
        ("nose", "nose"),
        ("docopt == 0.6.1             # Version Matching. Must be version 0.6.1", "docopt==0.6.1"),
        ("keyring >= 4.1.1            # Minimum version 4.1.1", "keyring>=4.1.1"),
        ("coverage != 3.5             # Version Exclusion. Anything except version 3.5", "coverage!=3.5"),
        ("Mopidy-Dirble ~= 1.1        # Compatible release. Same as >= 1.1, == 1.*", "Mopidy-Dirble~=1.1"),
        ("b==1.3", "b==1.3"),
        ("c >=1.2,<2.0", "c<2.0,>=1.2"),
        ("d[foo, bar]", "d[bar,foo]"),
        ("d[foo,  bar]", "d[bar,foo]"),
        ("d[bar,foo]", "d[bar,foo]"),
        ("e~=1.4.2", "e~=1.4.2"),
        ("f ==5.4 ; python_version < '2.7'", 'f==5.4; python_version < "2.7"'),
        ("g; sys_platform == 'win32'", 'g; sys_platform == "win32"'),
        (
            "http://w.org/w_P-3.0.3.dev1820+49a8884-cp34-none-win_amd64.whl",
            "http://w.org/w_P-3.0.3.dev1820+49a8884-cp34-none-win_amd64.whl",
        ),
        ("git+https://git.example.com/MyProject#egg=MyProject", "git+https://git.example.com/MyProject#egg=MyProject"),
        ("git+ssh://git.example.com/MyProject#egg=MyProject", "git+ssh://git.example.com/MyProject#egg=MyProject"),
        (
            "git+https://git.example.com/MyProject.git@da39a3ee5e6b4b0d3255bfef95601890afd80709#egg=MyProject",
            "git+https://git.example.com/MyProject.git@da39a3ee5e6b4b0d3255bfef95601890afd80709#egg=MyProject",
        ),
    ],
)
def test_requirements_txt(tmp_path: Path, req: str, key: str) -> None:
    req_file = RequirementsFile(req, root=tmp_path)
    expanded = req_file.validate_and_expand()
    if key:
        assert len(expanded) == 1
        assert expanded[0] == key
    else:
        assert expanded == []
    with req_file.with_file() as filename:
        assert filename.read_text() == req


def test_requirements_txt_local_path_file_protocol(tmp_path: Path) -> None:
    (tmp_path / "downloads").mkdir()
    (tmp_path / "downloads" / "numpy-1.9.2-cp34-none-win32.whl").write_text("1")
    raw = "numpy @ file://./downloads/numpy-1.9.2-cp34-none-win32.whl"
    req = RequirementsFile(raw, root=tmp_path)
    assert req.validate_and_expand() == ["numpy@ file://./downloads/numpy-1.9.2-cp34-none-win32.whl"]
    with req.with_file() as filename:
        assert filename.read_text() == raw


def test_requirements_txt_local_path_implicit(tmp_path: Path) -> None:
    (tmp_path / "downloads").mkdir()
    (tmp_path / "downloads" / "numpy-1.9.2-cp34-none-win32.whl").write_text("1")
    raw = "./downloads/numpy-1.9.2-cp34-none-win32.whl"
    req = RequirementsFile(raw, root=tmp_path)
    assert req.validate_and_expand() == [str(tmp_path / raw)]
    with req.with_file() as filename:
        assert filename.read_text() == raw


def test_requirements_env_var_present(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setenv("ENV_VAR", "beta")
    req = RequirementsFile("${ENV_VAR} >= 1", root=tmp_path)
    assert req.validate_and_expand() == ["beta>=1"]
    with req.with_file() as filename:
        assert filename.read_text() == "${ENV_VAR} >= 1"


def test_requirements_env_var_missing(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.delenv("ENV_VAR", raising=False)
    req = RequirementsFile("${ENV_VAR}", root=tmp_path)
    assert req.validate_and_expand() == []
    with req.with_file() as filename:
        assert filename.read_text() == "${ENV_VAR}"


def test_requirements_txt_transitive(tmp_path: Path) -> None:
    other_req = tmp_path / "other-requirements.txt"
    other_req.write_text("magic\nmagical")
    req = RequirementsFile("-r other-requirements.txt", root=tmp_path)
    assert req.validate_and_expand() == ["magic", "magical"]
    with req.with_file() as filename:
        assert filename.read_text() == "-r other-requirements.txt"


@pytest.mark.parametrize(
    "raw",
    [
        "--pre something",
        "-r one two",
        "--missing",
        "-k",
        "magic+https://git.example.com/MyProject#egg=MyProject",
    ],
)
def test_bad_line(raw: str) -> None:
    req = RequirementsFile(raw)
    with pytest.raises(ValueError, match=re.escape(raw)):
        req.validate_and_expand()



"""
Provides configuration values from tox.ini files.
"""
import logging
import os
from configparser import ConfigParser
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Type

from appdirs import user_config_dir

from tox.config.loader.ini import IniLoader

DEFAULT_CONFIG_FILE = Path(user_config_dir("tox")) / "config.ini"


class IniConfig:
    TOX_CONFIG_FILE_ENV_VAR = "TOX_CONFIG_FILE"
    STATE = {None: "failed to parse", True: "active", False: "missing"}

    section = "tox"

    def __init__(self) -> None:
        config_file = os.environ.get(self.TOX_CONFIG_FILE_ENV_VAR, None)
        self.is_env_var = config_file is not None
        self.config_file = Path(config_file if config_file is not None else DEFAULT_CONFIG_FILE)
        self._cache: Dict[Tuple[str, Type[Any]], Any] = {}
        self.has_config_file: Optional[bool] = self.config_file.exists()
        self.ini: Optional[IniLoader] = None

        if self.has_config_file:
            self.config_file = self.config_file.absolute()
            try:

                parser = ConfigParser()
                with self.config_file.open() as file_handler:
                    parser.read_file(file_handler)
                self.has_tox_section = parser.has_section("tox")
                self.ini = IniLoader("tox", parser, overrides=[]) if self.has_tox_section else None
            except Exception as exception:
                logging.error("failed to read config file %s because %r", config_file, exception)
                self.has_config_file = None

    def get(self, key: str, of_type: Type[Any]) -> Any:
        cache_key = key, of_type
        if cache_key in self._cache:
            result = self._cache[cache_key]
        else:
            try:
                if self.ini is None:  # pragma: no cover # this can only happen if we don't call __bool__ firsts
                    result = None
                else:
                    source = "file"
                    value = self.ini.load(key, of_type=of_type, conf=None, env_name="tox")
                    result = value, source
            except KeyError:  # just not found
                result = None
            except Exception as exception:  # noqa
                logging.warning("%s key %s as type %r failed with %r", self.config_file, key, of_type, exception)
                result = None
        self._cache[cache_key] = result
        return result

    def __bool__(self) -> bool:
        return bool(self.has_config_file) and bool(self.has_tox_section)

    @property
    def epilog(self) -> str:
        # text to show within the parsers epilog
        return (
            f"{os.linesep}config file {str(self.config_file)!r} {self.STATE[self.has_config_file]} "
            f"(change{'d' if self.is_env_var else ''} via env var {self.TOX_CONFIG_FILE_ENV_VAR})"
        )



"""Common functionality shared across multiple type of runs"""
import logging
import os
import time
from argparse import Action, ArgumentParser, ArgumentTypeError, Namespace
from concurrent.futures import CancelledError, Future, ThreadPoolExecutor, as_completed
from pathlib import Path
from signal import SIGINT, Handlers, signal
from threading import Event, Thread
from typing import Any, Dict, Iterator, List, Optional, Sequence, Set, Tuple, Union, cast

from colorama import Fore

from tox.config.types import EnvList
from tox.execute import Outcome
from tox.journal import write_journal
from tox.session.cmd.run.single import ToxEnvRunResult, run_one
from tox.session.state import State
from tox.tox_env.api import ToxEnv
from tox.tox_env.runner import RunToxEnv
from tox.util.graph import stable_topological_sort
from tox.util.spinner import MISS_DURATION, Spinner


class SkipMissingInterpreterAction(Action):
    def __call__(
        self,
        parser: ArgumentParser,  # noqa
        args: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        value = "true" if values is None else values
        if value not in ("config", "true", "false"):
            raise ArgumentTypeError(f"value must be 'config', 'true', or 'false' (got {repr(value)})")
        setattr(args, self.dest, value)


def env_run_create_flags(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--result-json",
        dest="result_json",
        metavar="path",
        of_type=Path,
        default=None,
        help="write a json file with detailed information about all commands and results involved",
    )
    parser.add_argument(
        "-s",
        "--skip-missing-interpreters",
        default="config",
        metavar="v",
        nargs="?",
        action=SkipMissingInterpreterAction,
        help="don't fail tests for missing interpreters: {config,true,false} choice",
    )
    parser.add_argument(
        "-r",
        "--recreate",
        dest="recreate",
        help="recreate the tox environments",
        action="store_true",
    )
    parser.add_argument(
        "-n",
        "--notest",
        dest="no_test",
        help="do not run the test commands",
        action="store_true",
    )
    parser.add_argument(
        "-b",
        "--pkg-only",
        "--sdistonly",
        action="store_true",
        help="only perform the packaging activity",
        dest="package_only",
    )
    parser.add_argument(
        "--installpkg",
        help="use specified package for installation into venv, instead of creating an sdist.",
        default=None,
        of_type=Path,
    )
    parser.add_argument(
        "--develop",
        action="store_true",
        help="install package in develop mode",
        dest="develop",
    )
    parser.add_argument(
        "--hashseed",
        metavar="SEED",
        help="set PYTHONHASHSEED to SEED before running commands. Defaults to a random integer in the range "
        "[1, 4294967295] ([1, 1024] on Windows). Passing 'noset' suppresses this behavior.",
        type=str,
        default="noset",
    )
    parser.add_argument(
        "--discover",
        dest="discover",
        nargs="+",
        metavar="path",
        help="for python discovery first try the python executables under these paths",
        default=[],
    )
    parser.add_argument(
        "--no-recreate-pkg",
        dest="no_recreate_pkg",
        help="if recreate is set do not recreate packaging tox environment(s)",
        action="store_true",
    )
    parser.add_argument(
        "--skip-pkg-install",
        dest="skip_pkg_install",
        help="skip package installation for this run",
        action="store_true",
    )


def run_and_report(state: State, result: List[ToxEnvRunResult]) -> int:
    # manifest the results
    name_to_run = {r.name: r for r in result}
    runs = [name_to_run[n] for n in list(state.env_list(everything=False))]
    # write the journal
    write_journal(getattr(state.options, "result_json", None), state.journal)
    # report the outcome
    return report(state.options.start, runs, state.options.is_colored)


def report(start: float, runs: List[ToxEnvRunResult], is_colored: bool) -> int:
    def _print(color: int, message: str) -> None:
        print(f"{color if is_colored else ''}{message}{Fore.RESET if is_colored else ''}")

    end = time.monotonic()
    all_ok = True
    for run in runs:
        ok = run.code == Outcome.OK
        msg = ("SKIP" if run.skipped else "OK") if ok else f"FAIL code {run.code}"
        duration_individual = [o.elapsed for o in run.outcomes]
        extra = f"+cmd[{','.join(f'{i:.2f}' for i in duration_individual)}]" if len(duration_individual) else ""
        setup = run.duration - sum(duration_individual)
        out = f"  {run.name}: {msg} ({run.duration:.2f}{f'=setup[{setup:.2f}]{extra}' if extra else ''} seconds)"
        _print((Fore.YELLOW if run.skipped else Fore.GREEN) if ok else Fore.RED, out)
        all_ok = ok and all_ok
    duration = end - start
    if all_ok:
        _print(Fore.GREEN, f"  congratulations :) ({duration:.2f} seconds)")
        return Outcome.OK
    else:
        _print(Fore.RED, f"  evaluation failed :( ({duration:.2f} seconds)")
        return runs[0].code if len(runs) == 1 else -1


logger = logging.getLogger(__name__)


def execute(state: State, max_workers: Optional[int], spinner: bool, live: bool) -> int:
    interrupt, done = Event(), Event()
    results: List[ToxEnvRunResult] = []
    future_to_env: Dict["Future[ToxEnvRunResult]", ToxEnv] = {}
    for env in list(state.env_list()):  # ensure envs can be constructed
        state.tox_env(env)
    previous, has_previous = None, False
    try:
        try:
            thread = Thread(
                target=_queue_and_wait,
                name="tox-interrupt",
                args=(state, results, future_to_env, interrupt, done, max_workers, spinner, live),
            )
            thread.start()
            thread.join()
        except KeyboardInterrupt:
            previous, has_previous = signal(SIGINT, Handlers.SIG_IGN), True
            logger.error(f"[{os.getpid()}] KeyboardInterrupt - teardown started")
            interrupt.set()
            for future, tox_env in list(future_to_env.items()):
                if future.cancel() is False and not future.done():  # if cannot be cancelled and not done -> still runs
                    tox_env.interrupt()
            done.wait()
    finally:
        exit_code = run_and_report(state, results)
        if has_previous:
            signal(SIGINT, previous)
        if "_TOX_SHOW_THREAD" in os.environ:  # pragma: no cover
            import threading  # pragma: no cover

            for thread in threading.enumerate():  # pragma: no cover
                print(thread)  # pragma: no cover
    return exit_code


class ToxSpinner(Spinner):
    def __init__(self, enabled: bool, state: State, total: int) -> None:
        super().__init__(
            enabled=enabled,
            colored=state.options.is_colored,
            stream=state.log_handler.stdout,
            total=total,
        )

    def update_spinner(self, result: ToxEnvRunResult, success: bool) -> None:
        if success:
            if result.skipped:
                done = self.skip
            else:
                done = self.succeed
        else:
            done = self.fail
        done(result.name)


def _queue_and_wait(
    state: State,
    results: List[ToxEnvRunResult],
    future_to_env: Dict["Future[ToxEnvRunResult]", ToxEnv],
    interrupt: Event,
    done: Event,
    max_workers: Optional[int],
    has_spinner: bool,
    live: bool,
) -> None:
    try:
        options = state.options
        to_run_list = list(state.env_list())
        with ToxSpinner(has_spinner, state, len(to_run_list)) as spinner:
            max_workers = len(to_run_list) if max_workers is None else max_workers
            completed: Set[str] = set()
            envs_to_run_generator = ready_to_run_envs(state, to_run_list, completed)

            def _run(tox_env: RunToxEnv) -> ToxEnvRunResult:
                spinner.add(cast(str, tox_env.conf.name))
                return run_one(tox_env, options.recreate, options.no_test, suspend_display=live is False)

            try:
                executor = ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix="tox-driver")
                while True:
                    env_list: List[str] = next(envs_to_run_generator, [])
                    if not env_list and not future_to_env:
                        break

                    for env in env_list:  # queue all available
                        tox_env_to_run = state.tox_env(env)
                        if interrupt.is_set():  # queue the rest as failed upfront
                            tox_env_to_run.teardown()
                            future: "Future[ToxEnvRunResult]" = Future()
                            res = ToxEnvRunResult(name=env, skipped=False, code=-2, outcomes=[], duration=MISS_DURATION)
                            future.set_result(res)
                        else:
                            future = executor.submit(_run, tox_env_to_run)
                        future_to_env[future] = tox_env_to_run
                    future = next(as_completed(future_to_env))
                    tox_env_done = future_to_env.pop(future)
                    try:
                        result: ToxEnvRunResult = future.result()
                    except CancelledError:
                        tox_env_done.teardown()
                        name = cast(str, tox_env_done.conf.name)
                        result = ToxEnvRunResult(name=name, skipped=False, code=-3, outcomes=[], duration=MISS_DURATION)
                    results.append(result)
                    completed.add(result.name)
                    _handle_one_run_done(result, spinner, state, live)
            except BaseException:  # pragma: no cover # noqa
                logging.exception("Internal Error")  # pragma: no cover
                raise  # pragma: no cover
            finally:
                executor.shutdown(wait=True)
    finally:
        done.set()


def _handle_one_run_done(result: ToxEnvRunResult, spinner: ToxSpinner, state: State, live: bool) -> None:
    success = result.code == Outcome.OK
    spinner.update_spinner(result, success)
    if live is False and state.options.parallel_live is False:  # teardown background run
        tox_env = state.tox_env(result.name)
        out_err = tox_env.close_and_read_out_err()  # sync writes from buffer to stdout/stderr
        has_package = tox_env.package_env is not None
        pkg_out_err = tox_env.package_env.close_and_read_out_err() if has_package else None  # type: ignore
        if not success or tox_env.conf["parallel_show_output"]:
            if pkg_out_err is not None:  # pragma: no branch # first show package build
                state.log_handler.write_out_err(pkg_out_err)  # pragma: no cover
            if out_err is not None:  # pragma: no branch # first show package build
                state.log_handler.write_out_err(out_err)


def ready_to_run_envs(state: State, to_run: List[str], completed: Set[str]) -> Iterator[List[str]]:
    """Generate tox environments ready to run"""
    order, todo = run_order(state, to_run)
    at = 0
    while order:
        ready_to_run: List[str] = []
        new_order: List[str] = []
        for env in order:  # collect next batch of ready to run
            if todo[env] - completed:
                new_order.append(env)
            else:
                ready_to_run.append(env)
            at += 1
        order = new_order
        yield ready_to_run


def run_order(state: State, to_run: List[str]) -> Tuple[List[str], Dict[str, Set[str]]]:
    to_run_set = set(to_run)
    todo: Dict[str, Set[str]] = {
        env: (to_run_set & set(cast(EnvList, state.tox_env(env).conf["depends"]).envs)) for env in to_run
    }
    order = stable_topological_sort(todo)
    return order, todo



from configparser import ConfigParser, SectionProxy
from typing import List, Optional, Set, TypeVar

from tox.config.loader.api import Loader, Override
from tox.config.loader.ini.factor import filter_for_env
from tox.config.loader.ini.replace import replace
from tox.config.loader.str_convert import StrConvert
from tox.config.main import Config
from tox.report import HandledError

V = TypeVar("V")


class IniLoader(StrConvert, Loader[str]):
    """Load configuration from an ini section (ini file is a string to string dictionary)"""

    def __init__(
        self,
        section: str,
        parser: ConfigParser,
        overrides: List[Override],
    ) -> None:
        self._section: SectionProxy = parser[section]
        self._parser = parser
        super().__init__(overrides)

    def load_raw(self, key: str, conf: Optional[Config], env_name: Optional[str]) -> str:
        value = self._section[key]
        collapsed_newlines = value.replace("\\\r\n", "").replace("\\\n", "")  # collapse explicit new-line escape
        if conf is None:  # conf is None when we're loading the global tox configuration file for the CLI
            replaced = collapsed_newlines  # we don't support factor and replace functionality there
        else:
            factor_selected = filter_for_env(collapsed_newlines, env_name)  # select matching factors
            try:
                replaced = replace(factor_selected, conf, env_name, self)  # do replacements
            except Exception as exception:
                msg = f"replace failed in {'tox' if env_name is None else env_name}.{key} with {exception!r}"
                raise HandledError(msg)
        return replaced

    def found_keys(self) -> Set[str]:
        return set(self._section.keys())

    def get_section(self, name: str) -> Optional[SectionProxy]:
        # needed for non tox environment replacements
        if self._parser.has_section(name):
            return self._parser[name]
        return None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(section={self._section}, overrides={self.overrides!r})"



"""
This module pulls together this package: create and parse CLI arguments for tox.
"""

from typing import Dict, Optional, Sequence, Tuple, cast

from tox.report import ToxHandler, setup_report

from .parser import Handler, Parsed, ToxParser

Handlers = Dict[str, Handler]


def get_options(*args: str) -> Tuple[Parsed, Handlers, Optional[Sequence[str]], ToxHandler]:
    pos_args: Optional[Tuple[str, ...]] = None
    try:  # remove positional arguments passed to parser if specified, they are pulled directly from sys.argv
        pos_arg_at = args.index("--")
        pos_args = tuple(args[pos_arg_at + 1 :])
        args = args[:pos_arg_at]
    except ValueError:
        pass

    guess_verbosity, log_handler = _get_base(args)
    parsed, cmd_handlers = _get_all(args)
    if guess_verbosity != parsed.verbosity:
        setup_report(parsed.verbosity, parsed.is_colored)  # pragma: no cover
    return parsed, cmd_handlers, pos_args, log_handler


def _get_base(args: Sequence[str]) -> Tuple[int, ToxHandler]:
    """First just load the base options (verbosity+color) to setup the logging framework."""
    tox_parser = ToxParser.base()
    parsed, _ = tox_parser.parse_known_args(args)
    guess_verbosity = parsed.verbosity
    handler = setup_report(guess_verbosity, parsed.is_colored)
    return guess_verbosity, handler


def _get_all(args: Sequence[str]) -> Tuple[Parsed, Handlers]:
    """Parse all the options."""
    tox_parser = _get_parser()
    parsed = cast(Parsed, tox_parser.parse_args(args))
    handlers = {k: p for k, (_, p) in tox_parser.handlers.items()}
    return parsed, handlers


def _get_parser() -> ToxParser:
    tox_parser = ToxParser.core()  # load the core options
    # plus options setup by plugins
    from tox.plugin.manager import MANAGER  # noqa

    MANAGER.tox_add_option(tox_parser)
    tox_parser.fix_defaults()
    return tox_parser


__all__ = (
    "get_options",
    "Handlers",
)



"""
Apply value substitution (replacement) on tox strings.
"""
import os
import re
import sys
from configparser import SectionProxy
from typing import TYPE_CHECKING, Iterator, List, Optional, Sequence, Tuple, Union

from tox.config.loader.stringify import stringify
from tox.config.main import Config
from tox.config.sets import ConfigSet
from tox.execute.request import shell_cmd

if TYPE_CHECKING:
    from tox.config.loader.ini import IniLoader

CORE_PREFIX = "tox"
BASE_TEST_ENV = "testenv"

ARGS_GROUP = re.compile(r"(?<!\\):")


def replace(value: str, conf: Config, name: Optional[str], loader: "IniLoader") -> str:
    # perform all non-escaped replaces
    start, end = 0, 0
    while True:
        start, end, match = _find_replace_part(value, start, end)
        if not match:
            break
        to_replace = value[start + 1 : end]
        replaced = _replace_match(conf, name, loader, to_replace)
        if replaced is None:
            # if we cannot replace, keep what was there, and continue looking for additional replaces following
            # note, here we cannot raise because the content may be a factorial expression, and in those case we don't
            # want to enforce escaping curly braces, e.g. it should work to write: env_list = {py39,py38}-{,dep}
            start = end = end + 1
            continue
        new_value = value[:start] + replaced + value[end + 1 :]
        start, end = 0, 0  # if we performed a replace start over
        if new_value == value:  # if we're not making progress stop (circular reference?)
            break
        value = new_value
    # remove escape sequences
    value = value.replace("\\{", "{")
    value = value.replace("\\}", "}")
    return value


def _find_replace_part(value: str, start: int, end: int) -> Tuple[int, int, bool]:
    match = False
    while end != -1:
        end = value.find("}", end)
        if end == -1:
            continue
        if end >= 1 and value[end - 1] == "\\":  # ignore escaped
            end += 1
            continue
        before = end
        while True:
            start = value.rfind("{", 0, before)
            if start >= 1 and value[start - 1] == "\\":  # ignore escaped
                before = start - 1
                continue
            match = start != -1
            break
        break  # pragma: no cover # for some odd reason this line is not reported by coverage, though is needed
    return start, end, match


def _replace_match(conf: Config, current_env: Optional[str], loader: "IniLoader", value: str) -> Optional[str]:
    of_type, *args = ARGS_GROUP.split(value)
    if of_type == "env":
        replace_value: Optional[str] = replace_env(args)
    elif of_type == "tty":
        replace_value = replace_tty(args)
    elif of_type == "posargs":
        replace_value = replace_pos_args(args, conf.pos_args)
    else:
        replace_value = replace_reference(conf, current_env, loader, value)
    return replace_value


_REPLACE_REF = re.compile(
    rf"""
    (\[(?P<full_env>{BASE_TEST_ENV}(:(?P<env>[^]]+))?|(?P<section>\w+))\])? # env/section
    (?P<key>[a-zA-Z0-9_]+) # key
    (:(?P<default>.*))? # default value
""",
    re.VERBOSE,
)


def replace_reference(
    conf: Config,
    current_env: Optional[str],
    loader: "IniLoader",
    value: str,
) -> Optional[str]:
    # a return value of None indicates could not replace
    match = _REPLACE_REF.match(value)
    if match:
        settings = match.groupdict()

        key = settings["key"]
        if settings["section"] is None and settings["full_env"] == BASE_TEST_ENV:
            settings["section"] = BASE_TEST_ENV

        exception: Optional[Exception] = None
        try:
            for src in _config_value_sources(settings["env"], settings["section"], current_env, conf, loader):
                try:
                    if isinstance(src, SectionProxy):
                        return src[key]
                    value = src[key]
                    as_str, _ = stringify(value)
                    return as_str
                except KeyError as exc:  # if fails, keep trying maybe another source can satisfy
                    exception = exc
        except Exception as exc:
            exception = exc
        if exception is not None:
            if isinstance(exception, KeyError):  # if the lookup failed replace - else keep
                default = settings["default"]
                if default is not None:
                    return default
                # we cannot raise here as that would mean users could not write factorials: depends = {py39,py38}-{,b}
            else:
                raise exception
    return None


def _config_value_sources(
    env: Optional[str],
    section: Optional[str],
    current_env: Optional[str],
    conf: Config,
    loader: "IniLoader",
) -> Iterator[Union[SectionProxy, ConfigSet]]:
    # if we have an env name specified take only from there
    # config is None only when loading the global tox config file for the CLI arguments, in this case no replace works
    if env is not None:
        if env in conf:
            yield conf.get_env(env)
        else:
            raise KeyError(f"missing tox environment with name {env}")

    # if we have a section name specified take only from there
    if section is not None:
        # special handle the core section under name tox
        if section == CORE_PREFIX:
            yield conf.core  # try via registered configs
        value = loader.get_section(section)  # fallback to section
        if value is not None:
            yield value
        return

    # otherwise try first from core conf, and fallback to our own environment
    yield conf.core
    if current_env is not None:
        yield conf.get_env(current_env)


def replace_pos_args(args: List[str], pos_args: Optional[Sequence[str]]) -> str:
    if pos_args is None:
        if args:
            replace_value = ":".join(args)  # if we use the defaults join back remaining args
        else:
            replace_value = ""
    else:
        replace_value = shell_cmd(pos_args)
    return replace_value


def replace_env(args: List[str]) -> str:
    key = args[0]
    default = "" if len(args) == 1 else args[1]
    return os.environ.get(key, default)


def replace_tty(args: List[str]) -> str:
    if sys.stdout.isatty():
        result = args[0] if len(args) > 0 else ""
    else:
        result = args[1] if len(args) > 1 else ""
    return result


__all__ = (
    "CORE_PREFIX",
    "BASE_TEST_ENV",
    "replace",
)



import subprocess
import sys
from datetime import date, datetime
from pathlib import Path

import sphinx_rtd_theme

from tox.version import __version__

company = "tox-dev"
name = "tox"
version = ".".join(__version__.split(".")[:2])
release = __version__
copyright = f"2010-{date.today().year}, {company}"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.extlinks",
]

templates_path = []
unused_docs = []
source_suffix = ".rst"
exclude_patterns = ["_build", "changelog/*", "_draft.rst"]

master_doc = "index"
pygments_style = "default"
always_document_param_types = True
project = name
today_fmt = "%B %d, %Y"

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_theme_options = {
    "canonical_url": "https://tox.readthedocs.io/en/latest/",
    "logo_only": False,
    "display_version": True,
    "prev_next_buttons_location": "bottom",
    "collapse_navigation": False,
    "sticky_navigation": True,
    "navigation_depth": 6,
    "includehidden": True,
}
html_static_path = ["_static"]
html_last_updated_fmt = datetime.now().isoformat()
html_logo = "_static/img/tox.svg"
html_favicon = "_static/img/toxfavi.ico"
htmlhelp_basename = "Pastedoc"
autoclass_content = "both"  # Include __init__ in class documentation
autodoc_member_order = "bysource"
autosectionlabel_prefix_document = True

extlinks = {
    "issue": ("https://github.com/tox-dev/tox/issues/%s", "#"),
    "pull": ("https://github.com/tox-dev/tox/pull/%s", "PR #"),
    "user": ("https://github.com/%s", "@"),
    "pypi": ("https://pypi.org/project/%s", ""),
}


def generate_draft_news():
    root = Path(__file__).parents[1]
    exe = Path(sys.executable)
    towncrier = exe.with_name(f"towncrier{exe.suffix}")
    new = subprocess.check_output([str(towncrier), "--draft", "--version", "NEXT"], cwd=root, universal_newlines=True)
    (root / "docs" / "_draft.rst").write_text("" if "No significant changes" in new else new)


generate_draft_news()


def setup(app):
    app.add_css_file("custom.css")



from collections import OrderedDict
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
)

from .of_type import ConfigConstantDefinition, ConfigDefinition, ConfigDynamicDefinition
from .types import EnvList

if TYPE_CHECKING:
    from tox.config.loader.api import Loader
    from tox.config.main import Config

V = TypeVar("V")


class ConfigSet:
    """A set of configuration that belong together (such as a tox environment settings, core tox settings)"""

    def __init__(self, conf: "Config", name: Optional[str]):
        self.name = name
        self._conf = conf
        self.loaders: List[Loader[Any]] = []
        self._defined: Dict[str, ConfigDefinition[Any]] = {}
        self._keys: Dict[str, None] = OrderedDict()

    def add_config(
        self,
        keys: Union[str, Sequence[str]],
        of_type: Type[V],
        default: Union[Callable[["Config", Optional[str]], V], V],
        desc: str,
        post_process: Optional[Callable[[V, "Config"], V]] = None,
        overwrite: bool = False,
    ) -> ConfigDynamicDefinition[V]:
        """
        Add configuration value.
        """
        keys_ = self._make_keys(keys)
        definition = ConfigDynamicDefinition(keys_, desc, self.name, of_type, default, post_process)
        result = self._add_conf(keys_, definition)
        return cast(ConfigDynamicDefinition[V], result)

    def add_constant(self, keys: Union[str, Sequence[str]], desc: str, value: V) -> ConfigConstantDefinition[V]:
        keys_ = self._make_keys(keys)
        definition = ConfigConstantDefinition(keys_, desc, self.name, value)
        result = self._add_conf(keys_, definition)
        return cast(ConfigConstantDefinition[V], result)

    @staticmethod
    def _make_keys(keys: Union[str, Sequence[str]]) -> Sequence[str]:
        return (keys,) if isinstance(keys, str) else keys

    def _add_conf(self, keys: Sequence[str], definition: ConfigDefinition[V]) -> ConfigDefinition[V]:
        key = keys[0]
        if key in self._defined:
            earlier = self._defined[key]
            # core definitions may be defined multiple times as long as all their options match, first defined wins
            if self.name is None and definition == earlier:
                definition = earlier
            else:
                raise ValueError(f"config {key} already defined")
        else:
            self._keys[key] = None
            for key in keys:
                self._defined[key] = definition
        return definition

    def __getitem__(self, item: str) -> Any:
        config_definition = self._defined[item]
        return config_definition(self._conf, item, self.loaders)

    def __repr__(self) -> str:
        values = (v for v in (f"name={self.name!r}" if self.name else "", f"loaders={self.loaders!r}") if v)
        return f"{self.__class__.__name__}({', '.join(values)})"

    def __iter__(self) -> Iterator[str]:
        return iter(self._keys.keys())

    def unused(self) -> List[str]:
        """Return a list of keys present in the config source but not used"""
        found: Set[str] = set()
        for loader in self.loaders:
            found.update(loader.found_keys())
        found -= self._defined.keys()
        return list(sorted(found))


class CoreConfigSet(ConfigSet):
    def __init__(self, conf: "Config", root: Path) -> None:
        super().__init__(conf, name=None)
        self.add_config(
            keys=["tox_root", "toxinidir"],
            of_type=Path,
            default=root,
            desc="the root directory (where the configuration file is found)",
        )
        work_dir_builder = (
            lambda conf, _: (conf.work_dir if conf.work_dir is not None else cast(Path, self["tox_root"])) / ".tox4"
        )
        self.add_config(
            keys=["work_dir", "toxworkdir"],
            of_type=Path,
            # here we pin to .tox4 to be able to use in parallel with v3 until final release
            default=work_dir_builder,
            desc="working directory",
        )
        self.add_config(
            keys=["temp_dir"],
            of_type=Path,
            default=lambda conf, _: cast(Path, self["tox_root"]) / ".temp",
            desc="temporary directory cleaned at start",
        )
        self.add_config(
            keys=["env_list", "envlist"],
            of_type=EnvList,
            default=EnvList([]),
            desc="define environments to automatically run",
        )



from pathlib import Path
from typing import Dict

import pytest

from tox.execute.request import ExecuteRequest, StdinSource


def test_execute_request_raise_on_empty_cmd(os_env: Dict[str, str]) -> None:
    with pytest.raises(ValueError, match="cannot execute an empty command"):
        ExecuteRequest(cmd=[], cwd=Path().absolute(), env=os_env, stdin=StdinSource.OFF, run_id="")



import json
import re
import sys
from pathlib import Path
from signal import SIGINT
from subprocess import PIPE, Popen
from time import sleep
from typing import Any, Dict, List, Tuple, Union

import pytest
from re_assert import Matches
from virtualenv.discovery.py_info import PythonInfo

from tox import __version__
from tox.pytest import ToxProjectCreator
from tox.tox_env.api import ToxEnv
from tox.tox_env.info import Info


def test_run_ignore_cmd_exit_code(tox_project: ToxProjectCreator) -> None:
    cmd = [
        "- python -c 'import sys; print(\"magic fail\", file=sys.stderr); sys.exit(1)'",
        "python -c 'import sys; print(\"magic pass\", file=sys.stdout); sys.exit(0)'",
    ]
    project = tox_project({"tox.ini": f"[tox]\nno_package=true\n[testenv]\ncommands={cmd[0]}\n {cmd[1]}"})
    outcome = project.run("r", "-e", "py")
    outcome.assert_success()
    assert "magic pass" in outcome.out
    assert "magic fail" in outcome.err


def test_run_sequential_fail(tox_project: ToxProjectCreator) -> None:
    def _cmd(value: int) -> str:
        return f"python -c 'import sys; print(\"exit {value}\"); sys.exit({value})'"

    ini = f"[tox]\nenv_list=a,b\nno_package=true\n[testenv:a]\ncommands={_cmd(1)}\n[testenv:b]\ncommands={_cmd(0)}"
    project = tox_project({"tox.ini": ini})
    outcome = project.run("r", "-e", "a,b")
    outcome.assert_failed()
    reports = outcome.out.splitlines()[-3:]
    assert Matches(r"  evaluation failed :\( \(.* seconds\)") == reports[-1]
    assert Matches(r"  b: OK \(.*=setup\[.*\]\+cmd\[.*\] seconds\)") == reports[-2]
    assert Matches(r"  a: FAIL code 1 \(.*=setup\[.*\]\+cmd\[.*\] seconds\)") == reports[-3]


@pytest.mark.timeout(120)
@pytest.mark.integration
def test_result_json_sequential(tox_project: ToxProjectCreator) -> None:
    cmd = [
        "- python -c 'import sys; print(\"magic fail\", file=sys.stderr); sys.exit(1)'",
        "python -c 'import sys; print(\"magic pass\"); sys.exit(0)'",
    ]
    project = tox_project(
        {
            "tox.ini": f"[tox]\nenvlist=py\n[testenv]\npackage=wheel\ncommands={cmd[0]}\n {cmd[1]}",
            "setup.py": "from setuptools import setup\nsetup(name='a', version='1.0', py_modules=['run'],"
            "install_requires=['setuptools>44'])",
            "run.py": "print('run')",
            "pyproject.toml": '[build-system]\nrequires=["setuptools","wheel"]\nbuild-backend="setuptools.build_meta"',
        }
    )
    log = project.path / "log.json"
    outcome = project.run("r", "-vv", "-e", "py", "--result-json", str(log))
    outcome.assert_success()
    with log.open("rt") as file_handler:
        log_report = json.load(file_handler)

    py_info = PythonInfo.current_system()
    host_python = {
        "executable": py_info.system_executable,
        "extra_version_info": None,
        "implementation": py_info.implementation,
        "is_64": py_info.architecture == 64,
        "sysplatform": py_info.platform,
        "version": py_info.version,
        "version_info": list(py_info.version_info),
    }
    packaging_setup = get_cmd_exit_run_id(log_report, ".package-py", "setup")

    assert packaging_setup == [
        (0, "install_requires"),
        (None, "get_requires_for_build_wheel"),
        (0, "install_build_requires"),
        (0, "freeze"),
        (None, "_exit"),
    ]
    packaging_test = get_cmd_exit_run_id(log_report, ".package-py", "test")
    assert packaging_test == [(None, "build_wheel")]
    packaging_installed = log_report["testenvs"][".package-py"].pop("installed_packages")
    assert {i[: i.find("==")] for i in packaging_installed} == {"pip", "setuptools", "wheel"}

    py_setup = get_cmd_exit_run_id(log_report, "py", "setup")
    assert py_setup == [(0, "install_package_deps"), (0, "install_package"), (0, "freeze")]
    py_test = get_cmd_exit_run_id(log_report, "py", "test")
    assert py_test == [(1, "commands[0]"), (0, "commands[1]")]
    packaging_installed = log_report["testenvs"]["py"].pop("installed_packages")
    expected_pkg = {"pip", "setuptools", "wheel", "a"}
    assert {i[: i.find("==")] if "@" not in i else "a" for i in packaging_installed} == expected_pkg
    install_package = log_report["testenvs"]["py"].pop("installpkg")
    assert re.match("^[a-fA-F0-9]{64}$", install_package.pop("sha256"))
    assert install_package == {"basename": "a-1.0-py3-none-any.whl", "type": "file"}

    expected = {
        "reportversion": "1",
        "toxversion": __version__,
        "platform": sys.platform,
        "testenvs": {
            "py": {"python": host_python},
            ".package-py": {"python": host_python},
        },
    }
    assert "host" in log_report
    assert log_report.pop("host")
    assert log_report == expected


def get_cmd_exit_run_id(report: Dict[str, Any], name: str, group: str) -> List[Tuple[Union[int, None], str]]:
    return [(i["retcode"], i["run_id"]) for i in report["testenvs"][name].pop(group)]


def test_rerun_sequential_skip(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=skip\ncommands=python -c 'print(1)'"})
    result_first = proj.run("--root", str(demo_pkg_inline))
    result_first.assert_success()
    result_rerun = proj.run("--root", str(demo_pkg_inline))
    result_rerun.assert_success()


def test_rerun_sequential_wheel(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    proj = tox_project(
        {"tox.ini": "[testenv]\npackage=wheel\ncommands=python -c 'from demo_pkg_inline import do; do()'"}
    )
    result_first = proj.run("--root", str(demo_pkg_inline))
    result_first.assert_success()
    result_rerun = proj.run("--root", str(demo_pkg_inline))
    result_rerun.assert_success()


@pytest.mark.integration
def test_rerun_sequential_sdist(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    proj = tox_project(
        {"tox.ini": "[testenv]\npackage=sdist\ncommands=python -c 'from demo_pkg_inline import do; do()'"}
    )
    result_first = proj.run("--root", str(demo_pkg_inline))
    result_first.assert_success()
    result_rerun = proj.run("--root", str(demo_pkg_inline))
    result_rerun.assert_success()


def test_recreate_package(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    proj = tox_project(
        {"tox.ini": "[testenv]\npackage=wheel\ncommands=python -c 'from demo_pkg_inline import do; do()'"}
    )
    result_first = proj.run("--root", str(demo_pkg_inline), "-r")
    result_first.assert_success()

    result_rerun = proj.run("-r", "--root", str(demo_pkg_inline), "--no-recreate-pkg")
    result_rerun.assert_success()


def test_package_deps_change(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    toml = (demo_pkg_inline / "pyproject.toml").read_text()
    build = (demo_pkg_inline / "build.py").read_text()
    ini = "[testenv]\npackage=wheel\ncommands=python -c 'from demo_pkg_inline import do; do()'"
    proj = tox_project({"tox.ini": ini, "pyproject.toml": toml, "build.py": build})

    result_first = proj.run("r")
    result_first.assert_success()
    assert ".package-py: install" not in result_first.out  # no deps initially

    # new deps are picked up
    (proj.path / "pyproject.toml").write_text(toml.replace("requires = []", 'requires = ["wheel"]'))
    (proj.path / "build.py").write_text(
        build.replace(
            "def get_requires_for_build_wheel(config_settings):\n    return []",
            "def get_requires_for_build_wheel(config_settings):\n    return ['setuptools']",
        )
    )

    result_rerun = proj.run("r")
    result_rerun.assert_success()

    # and installed
    rerun_install = [i for i in result_rerun.out.splitlines() if i.startswith(".package-py: install")]
    assert len(rerun_install) == 2
    assert rerun_install[0].endswith("wheel")
    assert rerun_install[1].endswith("setuptools")


@pytest.mark.skipif(sys.platform == "win32", reason="You need a conhost shell for keyboard interrupt")
def test_keyboard_interrupt(tox_project: ToxProjectCreator, demo_pkg_inline: Path, tmp_path: Path) -> None:
    marker = tmp_path / "a"
    proj = tox_project(
        {
            "tox.ini": "[testenv]\npackage=wheel\ncommands=python -c "
            f'\'from time import sleep; from pathlib import Path; p = Path("{str(marker)}"); p.write_text("");'
            " sleep(5)'\n"
            "[testenv:dep]\ndepends=py",
            "pyproject.toml": (demo_pkg_inline / "pyproject.toml").read_text(),
            "build.py": (demo_pkg_inline / "build.py").read_text(),
        }
    )
    cmd = [sys.executable, "-m", "tox", "-c", str(proj.path / "tox.ini"), "r", "-e", f"py,py{sys.version_info[0]},dep"]
    process = Popen(cmd, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    while not marker.exists():
        sleep(0.05)
    process.send_signal(SIGINT)
    out, err = process.communicate()
    assert process.returncode != 0
    assert "KeyboardInterrupt" in err, err
    assert "KeyboardInterrupt - teardown started\n" in out, out
    assert "interrupt tox environment: py\n" in out, out
    assert "requested interrupt of" in out, out
    assert "send signal SIGINT" in out, out
    assert "interrupt finished with success" in out, out
    assert "interrupt tox environment: .package-py" in out, out


def test_package_build_fails(tox_project: ToxProjectCreator) -> None:
    proj = tox_project(
        {
            "tox.ini": "[testenv]\npackage=wheel",
            "pyproject.toml": '[build-system]\nrequires=[]\nbuild-backend="build"\nbackend-path=["."]',
            "build.py": "",
        }
    )

    result = proj.run("r")
    result.assert_failed(code=1)
    assert "has no attribute 'build_wheel'" in result.out, result.out


def test_backend_not_found(tox_project: ToxProjectCreator) -> None:
    proj = tox_project(
        {
            "tox.ini": "[testenv]\npackage=wheel",
            "pyproject.toml": '[build-system]\nrequires=[]\nbuild-backend="build"',
            "build.py": "",
        }
    )

    result = proj.run("r")
    result.assert_failed(code=-5)
    assert "packaging backend failed (code=-5), with FailedToStart: could not start backend" in result.out, result.out


def test_missing_interpreter_skip_on(tox_project: ToxProjectCreator) -> None:
    ini = "[tox]\nskip_missing_interpreters=true\n[testenv]\npackage=skip\nbase_python=missing-interpreter"
    proj = tox_project({"tox.ini": ini})

    result = proj.run("r")
    result.assert_success()
    assert "py: SKIP" in result.out


def test_missing_interpreter_skip_off(tox_project: ToxProjectCreator) -> None:
    ini = "[tox]\nskip_missing_interpreters=false\n[testenv]\npackage=skip\nbase_python=missing-interpreter"
    proj = tox_project({"tox.ini": ini})

    result = proj.run("r")
    result.assert_failed()
    exp = "py: failed with could not find python interpreter matching any of the specs missing-interpreter"
    assert exp in result.out


def test_env_tmp_dir_reset(tox_project: ToxProjectCreator) -> None:
    ini = '[testenv]\npackage=skip\ncommands=python -c \'import os; os.mkdir(os.path.join( r"{env_tmp_dir}", "a"))\''
    proj = tox_project({"tox.ini": ini})
    result_first = proj.run("r")
    result_first.assert_success()

    result_second = proj.run("r", "-v", "-v")
    result_second.assert_success()
    assert "D clear env temp folder " in result_second.out, result_second.out


def test_env_name_change_recreate(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=skip\ncommands=\n"})
    result_first = proj.run("r")
    result_first.assert_success()

    tox_env = result_first.state.tox_env("py")
    assert repr(tox_env) == "VirtualEnvRunner(name=py)"
    path = tox_env.conf["env_dir"]
    with Info(path).compare({"name": "p", "type": "magical"}, ToxEnv.__name__):
        pass

    result_second = proj.run("r")
    result_second.assert_success()
    output = (
        "py: env type changed from {'name': 'p', 'type': 'magical'} to "
        "{'name': 'py', 'type': 'VirtualEnvRunner'}, will recreate"
    )
    assert output in result_second.out
    assert "py: remove tox env folder" in result_second.out


def test_deps_remove_recreate(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=skip\ndeps=wheel\n setuptools"})
    execute_calls = proj.patch_execute(lambda request: 0)
    result_first = proj.run("r")
    result_first.assert_success()
    assert execute_calls.call_count == 1

    (proj.path / "tox.ini").write_text("[testenv]\npackage=skip\ndeps=setuptools\n")
    result_second = proj.run("r")
    result_second.assert_success()
    assert "py: recreate env because dependencies removed: wheel" in result_second.out, result_second.out
    assert execute_calls.call_count == 2


def test_pkg_dep_remove_recreate(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    build = (demo_pkg_inline / "build.py").read_text()
    build_with_dep = build.replace("Summary: UNKNOWN\n", "Summary: UNKNOWN\n        Requires-Dist: wheel\n")
    proj = tox_project(
        {
            "tox.ini": "[testenv]\npackage=wheel",
            "pyproject.toml": (demo_pkg_inline / "pyproject.toml").read_text(),
            "build.py": build_with_dep,
        }
    )
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)

    result_first = proj.run("r")
    result_first.assert_success()
    run_ids = [i[0][2].run_id for i in execute_calls.call_args_list]
    assert run_ids == [
        "get_requires_for_build_wheel",
        "build_wheel",
        "install_package_deps",
        "install_package",
        "_exit",
    ]
    execute_calls.reset_mock()

    (proj.path / "build.py").write_text(build)
    result_second = proj.run("r")
    result_second.assert_success()
    assert "py: recreate env because dependencies removed: wheel" in result_second.out, result_second.out
    run_ids = [i[0][2].run_id for i in execute_calls.call_args_list]
    assert run_ids == ["get_requires_for_build_wheel", "build_wheel", "install_package", "_exit"]


def test_pkg_env_dep_remove_recreate(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    toml = (demo_pkg_inline / "pyproject.toml").read_text()
    proj = tox_project(
        {
            "tox.ini": "[testenv]\npackage=wheel",
            "pyproject.toml": toml.replace("requires = []", 'requires = ["setuptools"]'),
            "build.py": (demo_pkg_inline / "build.py").read_text(),
        }
    )
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result_first = proj.run("r")
    result_first.assert_success()
    run_ids = [i[0][2].run_id for i in execute_calls.call_args_list]
    assert run_ids == ["install_requires", "get_requires_for_build_wheel", "build_wheel", "install_package", "_exit"]
    execute_calls.reset_mock()

    (proj.path / "pyproject.toml").write_text(toml)
    result_second = proj.run("r")
    result_second.assert_success()
    assert ".package-py: recreate env because dependencies removed: setuptools" in result_second.out, result_second.out
    run_ids = [i[0][2].run_id for i in execute_calls.call_args_list]
    assert run_ids == ["get_requires_for_build_wheel", "build_wheel", "install_package", "_exit"]


def test_skip_pkg_install(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=wheel\n"})
    result_first = proj.run("--root", str(demo_pkg_inline), "--skip-pkg-install")
    result_first.assert_success()
    assert result_first.out.startswith("py: skip building and installing the package"), result_first.out



from collections import OrderedDict, defaultdict
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Sequence

from tox.config.loader.api import Loader, Override, OverrideMap
from tox.config.source import Source

from .sets import ConfigSet, CoreConfigSet


class Config:
    def __init__(
        self,
        config_source: Source,
        overrides: List[Override],
        root: Path,
        pos_args: Optional[Sequence[str]],
        work_dir: Path,
    ) -> None:
        self.pos_args = pos_args
        self.work_dir = work_dir
        self._root = root

        self._overrides: OverrideMap = defaultdict(list)
        for override in overrides:
            self._overrides[override.namespace].append(override)

        self._src = config_source
        self._env_to_set: Dict[str, ConfigSet] = OrderedDict()
        self._core_set: Optional[CoreConfigSet] = None
        self.register_config_set: Callable[[str, ConfigSet], Any] = lambda n, e: None

    @property
    def core(self) -> CoreConfigSet:
        if self._core_set is not None:
            return self._core_set
        core = CoreConfigSet(self, self._root)
        for loader in self._src.get_core(self._overrides):
            core.loaders.append(loader)

        from tox.plugin.manager import MANAGER

        MANAGER.tox_add_core_config(core)
        self._core_set = core
        return core

    def get_env(self, item: str, package: bool = False, loaders: Optional[Sequence[Loader[Any]]] = None) -> ConfigSet:
        try:
            return self._env_to_set[item]
        except KeyError:
            env = ConfigSet(self, item)
            self._env_to_set[item] = env
            if loaders is not None:
                env.loaders.extend(loaders)
            for loader in self._src.get_env_loaders(item, self._overrides, package, env):
                env.loaders.append(loader)
            # whenever we load a new configuration we need build a tox environment which process defines the valid
            # configuration values
            self.register_config_set(item, env)
            return env

    def __iter__(self) -> Iterator[str]:
        return self._src.envs(self.core)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(config_source={self._src!r})"

    def __contains__(self, item: str) -> bool:
        return any(name for name in self if name == item)



from typing import TYPE_CHECKING, Dict, Iterator, Optional, Sequence, Set, Tuple, cast

from tox.config.main import Config
from tox.config.sets import ConfigSet
from tox.journal import Journal
from tox.plugin.impl import impl
from tox.report import HandledError, ToxHandler
from tox.session.common import CliEnv
from tox.tox_env.package import PackageToxEnv
from tox.tox_env.runner import RunToxEnv

if TYPE_CHECKING:

    from tox.config.cli.parse import Handlers
    from tox.config.cli.parser import Parsed, ToxParser


class State:
    def __init__(
        self,
        conf: Config,
        opt_parse: Tuple["Parsed", "Handlers"],
        args: Sequence[str],
        log_handler: ToxHandler,
    ) -> None:
        self.conf = conf
        self.conf.register_config_set = self.register_config_set
        options, cmd_handlers = opt_parse
        self.options = options
        self.cmd_handlers = cmd_handlers
        self.log_handler = log_handler
        self.args = args

        self._run_env: Dict[str, RunToxEnv] = {}
        self._pkg_env: Dict[str, PackageToxEnv] = {}
        self._pkg_env_discovered: Set[str] = set()

        self.journal: Journal = Journal(getattr(options, "result_json", None) is not None)

    def env_list(self, everything: bool = False) -> Iterator[str]:
        fallback_env = "py"
        if everything:
            _at = 0
            for _at, env in enumerate(self.conf, start=1):
                yield env
            if _at == 0:  # if we discovered no other env, inject the default
                yield fallback_env
            return
        use_env_list: Optional[CliEnv] = getattr(self.options, "env", None)
        if use_env_list is None or use_env_list.all:
            use_env_list = self.conf.core["env_list"]
        if not use_env_list:
            use_env_list = CliEnv([fallback_env])
        if use_env_list is not None:  # pragma: no branch # can't happen
            yield from use_env_list

    def tox_env(self, name: str) -> RunToxEnv:
        if name in self._pkg_env_discovered:
            raise HandledError(f"cannot run packaging environment {name}")
        with self.log_handler.with_context(name):
            tox_env = self._run_env.get(name)
            if tox_env is not None:
                return tox_env
            self.conf.get_env(name)  # the lookup here will trigger register_config_set, which will build it
            return self._run_env[name]

    def register_config_set(self, name: str, config_set: ConfigSet) -> None:
        """Ensure the config set with the given name has been registered with configuration values"""
        # during the creation of hte tox environment we automatically register configurations, so to ensure
        # config sets have a set of defined values in it we have to ensure the tox environment is created
        if name in self._pkg_env_discovered:
            return  # packaging environments are created explicitly, nothing to do here
        if name in self._run_env:  # pragma: no branch
            raise ValueError(f"{name} run tox env already defined")  # pragma: no cover
        # runtime environments are created upon lookup via the tox_env method, call it
        self._build_run_env(config_set)

    def _build_run_env(self, env_conf: ConfigSet) -> None:
        env_conf.add_config(
            keys="runner",
            desc="the tox execute used to evaluate this environment",
            of_type=str,
            default=self.options.default_runner,  # noqa
        )
        runner = cast(str, env_conf["runner"])
        from tox.tox_env.register import REGISTER

        builder = REGISTER.runner(runner)
        name = cast(str, env_conf.name)
        journal = self.journal.get_env_journal(name)
        env: RunToxEnv = builder(env_conf, self.conf.core, self.options, journal, self.log_handler)
        self._run_env[name] = env
        self._build_package_env(env)

    def _build_package_env(self, env: RunToxEnv) -> None:
        pkg_env_gen = env.set_package_env()
        try:
            name, packager = next(pkg_env_gen)
        except StopIteration:
            pass
        else:
            with self.log_handler.with_context(name):
                package_tox_env = self._get_package_env(packager, name)
                try:
                    pkg_env_gen.send(package_tox_env)
                except StopIteration:
                    pass

    def _get_package_env(self, packager: str, name: str) -> PackageToxEnv:
        if name in self._pkg_env:  # if already created reuse
            pkg_tox_env: PackageToxEnv = self._pkg_env[name]
        else:
            from tox.tox_env.register import REGISTER

            package_type = REGISTER.package(packager)
            self._pkg_env_discovered.add(name)
            if name in self._run_env:
                raise HandledError(f"{name} is already defined as a run environment, cannot be packaging too")
            pkg_conf = self.conf.get_env(name, package=True)
            journal = self.journal.get_env_journal(name)
            pkg_tox_env = package_type(pkg_conf, self.conf.core, self.options, journal, self.log_handler)
            self._pkg_env[name] = pkg_tox_env
        return pkg_tox_env


@impl
def tox_add_option(parser: "ToxParser") -> None:
    from tox.tox_env.register import REGISTER

    parser.add_argument(
        "--runner",
        dest="default_runner",
        help="default execute",
        default=REGISTER.default_run_env,
        choices=list(REGISTER.run_envs),
    )



"""Sources."""
from abc import ABC, abstractmethod
from typing import Any, Iterator

from tox.config.loader.api import Loader, OverrideMap

from ..sets import ConfigSet, CoreConfigSet


class Source(ABC):
    """
    Source is able to return a configuration value (for either the core or per environment source).
    """

    @abstractmethod
    def get_core(self, override_map: OverrideMap) -> Iterator[Loader[Any]]:
        """Return the core loader from this source."""
        raise NotImplementedError

    @abstractmethod
    def get_env_loaders(
        self, env_name: str, override_map: OverrideMap, package: bool, conf: ConfigSet
    ) -> Iterator[Loader[Any]]:
        """Return the load for this environment."""
        raise NotImplementedError

    @abstractmethod
    def envs(self, core_conf: "CoreConfigSet") -> Iterator[str]:
        """Return a list of environments defined within this source"""
        raise NotImplementedError



import sys
from pathlib import Path
from typing import Callable, Optional, Sequence

import pytest
from _pytest.monkeypatch import MonkeyPatch

from tox.config.cli.parser import Parsed
from tox.config.loader.api import Override
from tox.config.main import Config
from tox.run import make_config

pytest_plugins = "tox.pytest"
HERE = Path(__file__).absolute().parent


@pytest.fixture(scope="session")
def value_error() -> Callable[[str], str]:
    def _fmt(msg: str) -> str:
        return f'ValueError("{msg}"{"," if sys.version_info < (3, 7) else ""})'

    return _fmt


if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Protocol
else:  # pragma: no cover (<py38)
    from typing_extensions import Protocol  # noqa


class ToxIniCreator(Protocol):
    def __call__(self, conf: str, override: Optional[Sequence[Override]] = None) -> Config:
        ...


@pytest.fixture
def tox_ini_conf(tmp_path: Path, monkeypatch: MonkeyPatch) -> ToxIniCreator:
    def func(conf: str, override: Optional[Sequence[Override]] = None) -> Config:
        dest = tmp_path / "c"
        dest.mkdir()
        config_file = dest / "tox.ini"
        config_file.write_bytes(conf.encode("utf-8"))
        with monkeypatch.context() as context:
            context.chdir(tmp_path)
        return make_config(
            Parsed(work_dir=dest, override=override or [], config_file=config_file, root_dir=None),
            pos_args=[],
        )

    return func


@pytest.fixture(scope="session")
def demo_pkg_setuptools() -> Path:
    return HERE / "demo_pkg_setuptools"


@pytest.fixture(scope="session")
def demo_pkg_inline() -> Path:
    return HERE / "demo_pkg_inline"



from tox.config.types import Command, EnvList


def tests_command_repr() -> None:
    cmd = Command(["python", "-m", "pip", "list"])
    assert repr(cmd) == "Command(args=['python', '-m', 'pip', 'list'])"
    assert cmd.ignore_exit_code is False


def tests_command_repr_ignore() -> None:
    cmd = Command(["-", "python", "-m", "pip", "list"])
    assert repr(cmd) == "Command(args=['-', 'python', '-m', 'pip', 'list'])"
    assert cmd.ignore_exit_code is True


def tests_command_eq() -> None:
    cmd_1 = Command(["python", "-m", "pip", "list"])
    cmd_2 = Command(["python", "-m", "pip", "list"])
    assert cmd_1 == cmd_2


def tests_command_ne() -> None:
    cmd_1 = Command(["python", "-m", "pip", "list"])
    cmd_2 = Command(["-", "python", "-m", "pip", "list"])
    assert cmd_1 != cmd_2


def tests_env_list_repr() -> None:
    env = EnvList(["py39", "py38"])
    assert repr(env) == "EnvList(['py39', 'py38'])"


def tests_env_list_eq() -> None:
    env_1 = EnvList(["py39", "py38"])
    env_2 = EnvList(["py39", "py38"])
    assert env_1 == env_2


def tests_env_list_ne() -> None:
    env_1 = EnvList(["py39", "py38"])
    env_2 = EnvList(["py38", "py39"])
    assert env_1 != env_2



from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from tox.pytest import ToxProjectCreator


def test_legacy_show_config(tox_project: ToxProjectCreator, mocker: MockerFixture) -> None:
    show_config = mocker.patch("tox.session.cmd.legacy.show_config")

    outcome = tox_project({"tox.ini": ""}).run("le", "--showconfig")

    assert show_config.call_count == 1
    assert outcome.state.options.list_keys_only == []
    assert outcome.state.options.show_core is True


@pytest.mark.parametrize("verbose", range(3))
def test_legacy_list_default(tox_project: ToxProjectCreator, mocker: MockerFixture, verbose: int) -> None:
    list_env = mocker.patch("tox.session.cmd.legacy.list_env")

    outcome = tox_project({"tox.ini": ""}).run("le", "-l", *(["-v"] * verbose))

    assert list_env.call_count == 1
    assert outcome.state.options.list_no_description is (verbose < 1)
    assert outcome.state.options.list_default_only is True
    assert outcome.state.options.show_core is False


@pytest.mark.parametrize("verbose", range(3))
def test_legacy_list_all(tox_project: ToxProjectCreator, mocker: MockerFixture, verbose: int) -> None:
    list_env = mocker.patch("tox.session.cmd.legacy.list_env")

    outcome = tox_project({"tox.ini": ""}).run("le", "-a", *(["-v"] * verbose))

    assert list_env.call_count == 1
    assert outcome.state.options.list_no_description is (verbose < 1)
    assert outcome.state.options.list_default_only is False
    assert outcome.state.options.show_core is False


def test_legacy_devenv(tox_project: ToxProjectCreator, mocker: MockerFixture, tmp_path: Path) -> None:
    devenv = mocker.patch("tox.session.cmd.legacy.devenv")
    into = tmp_path / "b"

    outcome = tox_project({"tox.ini": ""}).run("le", "--devenv", str(into), "-e", "py")

    assert devenv.call_count == 1
    assert outcome.state.options.devenv_path == into


def test_legacy_run_parallel(tox_project: ToxProjectCreator, mocker: MockerFixture) -> None:
    run_parallel = mocker.patch("tox.session.cmd.legacy.run_parallel")

    tox_project({"tox.ini": ""}).run("le", "-p", "all", "-e", "py")

    assert run_parallel.call_count == 1


def test_legacy_run_sequential(tox_project: ToxProjectCreator, mocker: MockerFixture) -> None:
    run_sequential = mocker.patch("tox.session.cmd.legacy.run_sequential")

    tox_project({"tox.ini": ""}).run("le", "-e", "py")

    assert run_sequential.call_count == 1



from abc import abstractmethod
from argparse import ArgumentTypeError
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Set, Type, TypeVar

from tox.plugin.impl import impl

from .convert import Convert
from .str_convert import StrConvert

if TYPE_CHECKING:
    from tox.config.cli.parser import ToxParser
    from tox.config.main import Config


class Override:
    def __init__(self, value: str) -> None:
        split_at = value.find("=")
        if split_at == -1:
            raise ArgumentTypeError(f"override {value} has no = sign in it")
        key = value[:split_at]
        ns_at = key.find(".")
        self.namespace = "" if ns_at == -1 else key[:ns_at]
        self.key = key[ns_at + 1 :]
        self.value = value[split_at + 1 :]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self}')"

    def __str__(self) -> str:
        return f"{self.namespace}{'.' if self.namespace else ''}{self.key}={self.value}"

    def __eq__(self, other: Any) -> bool:
        if type(self) != type(other):
            return False
        return (self.namespace, self.key, self.value) == (other.namespace, other.key, other.value)

    def __ne__(self, other: Any) -> bool:
        return not (self == other)


OverrideMap = Mapping[str, List[Override]]

T = TypeVar("T")
V = TypeVar("V")


class Loader(Convert[T]):
    """Loader loads a configuration value and converts it."""

    def __init__(self, overrides: List[Override]) -> None:
        self.overrides = {o.key: o for o in overrides}

    @abstractmethod
    def load_raw(self, key: str, conf: Optional["Config"], env_name: Optional[str]) -> T:
        """
        Load the raw object from the config store.

        :param key: the key under what we want the configuration
        :param env_name: the name of the environment this load is happening for
        :param conf: the global config object
        """
        raise NotImplementedError

    @abstractmethod
    def found_keys(self) -> Set[str]:
        """A list of configuration keys found within the configuration."""
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{type(self).__name__}"

    def load(self, key: str, of_type: Type[V], conf: Optional["Config"], env_name: Optional[str]) -> V:
        """
        Load a value.

        :param key: the key under it lives
        :param of_type: the type to convert to
        :param conf: the configuration object of this tox session (needed to manifest the value)
        :param env_name: env name
        :return: the converted type
        """
        if key in self.overrides:
            return _STR_CONVERT.to(self.overrides[key].value, of_type)
        raw = self.load_raw(key, conf, env_name)
        converted = self.to(raw, of_type)
        return converted


@impl
def tox_add_option(parser: "ToxParser") -> None:
    parser.add_argument(
        "-x",
        "--override",
        action="append",
        type=Override,
        default=[],
        dest="override",
        help="configuration override(s)",
    )


_STR_CONVERT = StrConvert()



"""Convert string configuration values to tox python configuration objects."""
import re
import shlex
import sys
from itertools import chain
from pathlib import Path
from typing import Any, Iterator, Tuple, Type

from tox.config.loader.convert import Convert
from tox.config.types import Command, EnvList


class StrConvert(Convert[str]):
    """A class converting string values to tox types"""

    @staticmethod
    def to_str(value: str) -> str:
        return str(value).strip()

    @staticmethod
    def to_path(value: str) -> Path:
        return Path(value)

    @staticmethod
    def to_list(value: str, of_type: Type[Any]) -> Iterator[str]:
        splitter = "\n" if issubclass(of_type, Command) or "\n" in value else ","
        splitter = splitter.replace("\r", "")
        for token in value.split(splitter):
            value = token.strip()
            if value:
                yield value

    @staticmethod
    def to_set(value: str, of_type: Type[Any]) -> Iterator[str]:
        for value in StrConvert.to_list(value, of_type):
            yield value

    @staticmethod
    def to_dict(value: str, of_type: Tuple[Type[Any], Type[Any]]) -> Iterator[Tuple[str, str]]:
        for row in value.split("\n"):
            row = row.strip()
            if row:
                try:
                    at = row.index("=")
                except ValueError as exc:
                    raise TypeError(f"dictionary lines must be of form key=value, found {row}") from exc
                else:
                    key = row[:at].strip()
                    value = row[at + 1 :].strip()
                    yield key, value

    @staticmethod
    def to_command(value: str) -> Command:
        win = sys.platform == "win32"
        splitter = shlex.shlex(value, posix=not win)
        splitter.whitespace_split = True
        if win:  # pragma: win32 cover
            args = []
            for arg in splitter:
                if arg[0] == "'" and arg[-1] == "'":  # remove outer quote - the arg is passed as one, so no need for it
                    arg = arg[1:-1]
                if "/" in arg:  # normalize posix paths to nt paths
                    arg = "\\".join(re.split(pattern=r"[\\/]", string=arg))
                args.append(arg)
        else:  # pragma: win32 no cover
            args = list(splitter)
        return Command(args)

    @staticmethod
    def to_env_list(value: str) -> EnvList:
        from tox.config.loader.ini.factor import extend_factors

        elements = list(chain.from_iterable(extend_factors(expr) for expr in value.split("\n")))
        return EnvList(elements)

    TRUTHFUL_VALUES = {"true", "1", "yes", "on"}
    FALSE_VALUES = {"false", "0", "no", "off", ""}
    VALID_BOOL = list(sorted(TRUTHFUL_VALUES | FALSE_VALUES))

    @staticmethod
    def to_bool(value: str) -> bool:
        norm = value.strip().lower()
        if norm in StrConvert.TRUTHFUL_VALUES:
            return True
        elif norm in StrConvert.FALSE_VALUES:
            return False
        else:
            raise TypeError(f"value {value} cannot be transformed to bool, valid: {', '.join(StrConvert.VALID_BOOL)}")


__all__ = ("StrConvert",)



import pytest

from tox.tox_env.python.virtual_env.runner import VirtualEnvRunner
from tox.tox_env.register import ToxEnvRegister


def test_register_set_new_default_no_register() -> None:
    register = ToxEnvRegister()
    with pytest.raises(ValueError, match="run env must be registered before setting it as default"):
        register.default_run_env = "new-env"


def test_register_set_new_default_with_register() -> None:
    register = ToxEnvRegister()
    register.add_run_env(VirtualEnvRunner)
    assert register.default_run_env == ""
    register.default_run_env = VirtualEnvRunner.id()
    assert register.default_run_env == "virtualenv"



import os
import platform
import re
import sys

import pytest
from packaging.version import Version
from pytest_mock import MockerFixture

from tox.config.types import Command
from tox.pytest import MonkeyPatch, ToxProjectCreator
from tox.version import __version__


def test_show_config_default_run_env(tox_project: ToxProjectCreator, monkeypatch: MonkeyPatch) -> None:
    py_ver = sys.version_info[0:2]
    name = "py{}{}".format(*py_ver) if platform.python_implementation() == "CPython" else "pypy3"
    project = tox_project({"tox.ini": f"[tox]\nenv_list = {name}\n[testenv:{name}]\ncommands={{posargs}}"})
    result = project.run("c", "--", "magic")
    state = result.state
    assert state.args == ("c", "--", "magic")
    outcome = list(state.env_list(everything=True))
    assert outcome == [name]

    path = re.escape(str(project.path))
    sep = re.escape(str(os.sep))
    version = re.escape(Version(__version__).public)

    monkeypatch.delenv("TERM", raising=False)  # disable conditionally set flag
    if sys.platform == "win32":  # pragma: win32 cover
        p_env = ["COMSPEC", "MSYSTEM", "PATHEXT", "PROCESSOR_ARCHITECTURE", "SYSTEMROOT", "TEMP", "TMP", "USERPROFILE"]
    else:  # pragma: win32 no cover
        p_env = ["TMPDIR"]
    p_env.extend(["PIP_*", "VIRTUALENV_*", "http_proxy", "https_proxy", "no_proxy"])
    pass_env_str = "\n".join(f"      {re.escape(p)}" for p in sorted(p_env))[4:]

    expected = rf"""
    \[testenv:{name}\]
    type = VirtualEnvRunner
    base = testenv
    runner = virtualenv
    env_name = {name}
    env_dir = {path}{sep}\.tox4{sep}{name}
    env_tmp_dir = {path}{sep}\.tox4{sep}{name}{sep}tmp
    set_env =
      PIP_DISABLE_PIP_VERSION_CHECK=1
      VIRTUALENV_NO_PERIODIC_UPDATE=1
    pass_env =
    {pass_env_str}\
    parallel_show_output = False
    description =
    commands = magic
    commands_pre =
    commands_post =
    change_dir = {path}
    depends =
    skip_install = False
    usedevelop = False
    package = sdist
    package_tox_env_type = virtualenv-pep-517-sdist
    package_env = \.package
    extras =
    base_python = {name}
    env_site_packages_dir = {path}{sep}\.tox4{sep}{name}{sep}.*
    env_python = {path}{sep}\.tox4{sep}{name}{sep}.*
    deps =\

    \[tox\]
    tox_root = {path}
    work_dir = {path}{sep}\.tox4
    temp_dir = {path}{sep}\.temp
    env_list = {name}
    min_version = {version}
    provision_tox_env = \.tox
    requires = tox>={version}
    no_package = False\
    skip_missing_interpreters = True
    """
    result.assert_out_err(expected, "", regex=True)


def test_show_config_commands(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.ini": """
        [tox]
        env_list = py
        no_package = true
        [testenv]
        commands_pre =
            python -c 'import sys; print("start", sys.executable)'
        commands =
            pip config list
            pip list
        commands_post =
            python -c 'import sys; print("end", sys.executable)'
        """,
        },
    )
    outcome = project.run("c")
    outcome.assert_success()
    env_config = outcome.state.tox_env("py").conf
    assert env_config["commands_pre"] == [Command(args=["python", "-c", 'import sys; print("start", sys.executable)'])]
    assert env_config["commands"] == [
        Command(args=["pip", "config", "list"]),
        Command(args=["pip", "list"]),
    ]
    assert env_config["commands_post"] == [Command(args=["python", "-c", 'import sys; print("end", sys.executable)'])]


def test_show_config_filter_keys(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv]\nmagic=yes"})
    outcome = project.run("c", "-e", "py", "-k", "no_package", "env_name", "--core")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\nenv_name = py\n\n[tox]\nno_package = False\n", "")


def test_show_config_unused(tox_project: ToxProjectCreator) -> None:
    outcome = tox_project({"tox.ini": "[testenv:py]\nmagical=yes\nmagic=yes"}).run("c", "-e", "py")
    outcome.assert_success()
    assert "\n# !!! unused: magic, magical\n" in outcome.out


def test_show_config_exception(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.ini": """
        [testenv:a]
        base_python = missing-python
        """,
        },
    )
    outcome = project.run("c", "-e", "a", "-k", "env_site_packages_dir")
    outcome.assert_success()
    txt = (
        "\nenv_site_packages_dir = # Exception: "
        "RuntimeError(\"failed to find interpreter for Builtin discover of python_spec='missing-python'"
    )
    assert txt in outcome.out


@pytest.mark.parametrize("stdout_is_atty", [True, False])
def test_pass_env_config_default(tox_project: ToxProjectCreator, stdout_is_atty: bool, mocker: MockerFixture) -> None:
    mocker.patch("sys.stdout.isatty", return_value=stdout_is_atty)
    project = tox_project({"tox.ini": ""})
    outcome = project.run("c", "-e", "py", "-k", "pass_env")
    pass_env = outcome.state.tox_env("py").conf["pass_env"]
    expected = (
        (["COMSPEC", "MSYSTEM", "PATHEXT"] if sys.platform == "win32" else [])
        + ["PIP_*"]
        + (["PROCESSOR_ARCHITECTURE", "SYSTEMROOT", "TEMP"] if sys.platform == "win32" else [])
        + (["TERM"] if stdout_is_atty else [])
        + (["TMP", "USERPROFILE"] if sys.platform == "win32" else ["TMPDIR"])
        + ["VIRTUALENV_*", "http_proxy", "https_proxy", "no_proxy"]
    )
    assert pass_env == expected



"""
A tox environment that can build packages.
"""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, List, Set

from packaging.requirements import Requirement

from tox.config.sets import ConfigSet
from tox.journal import EnvJournal
from tox.report import ToxHandler
from tox.util.threading import AtomicCounter

from .api import ToxEnv

if TYPE_CHECKING:
    from tox.config.cli.parser import Parsed


class PackageToxEnv(ToxEnv, ABC):
    def __init__(
        self, conf: ConfigSet, core: ConfigSet, options: "Parsed", journal: EnvJournal, log_handler: ToxHandler
    ) -> None:
        super().__init__(conf, core, options, journal, log_handler)
        self.recreate_package = options.no_recreate_pkg is False if options.recreate else False
        self.ref_count = AtomicCounter()

    @abstractmethod
    def get_package_dependencies(self, extras: Set[str]) -> List[Requirement]:
        raise NotImplementedError

    @abstractmethod
    def perform_packaging(self) -> List[Path]:
        raise NotImplementedError

    def clean(self, force: bool = False) -> None:
        if force or self.recreate_package:  # only recreate if user did not opt out
            super().clean(force)



"""Main entry point for tox."""
import logging
import sys
import time
from itertools import chain
from pathlib import Path
from typing import Optional, Sequence

from tox.config.cli.parse import get_options
from tox.config.cli.parser import Parsed
from tox.config.main import Config
from tox.config.source.tox_ini import ToxIni
from tox.provision import provision
from tox.report import HandledError, ToxHandler
from tox.session.state import State


def run(args: Optional[Sequence[str]] = None) -> None:
    try:
        with ToxHandler.patch_thread():
            result = main(sys.argv[1:] if args is None else args)
    except Exception as exception:
        if isinstance(exception, HandledError):
            logging.error("%s| %s", type(exception).__name__, str(exception))
            result = -2
        else:
            raise
    except KeyboardInterrupt:
        result = -2
    raise SystemExit(result)


def main(args: Sequence[str]) -> int:
    state = setup_state(args)
    result = provision(state)
    if result is not False:
        return result
    command = state.options.command
    handler = state.cmd_handlers[command]
    result = handler(state)
    return result


def setup_state(args: Sequence[str]) -> State:
    """Setup the state object of this run."""
    start = time.monotonic()
    # parse CLI arguments
    parsed, handlers, pos_args, log_handler = get_options(*args)
    parsed.start = start
    # parse configuration file
    config = make_config(parsed, pos_args)
    # build tox environment config objects
    state = State(config, (parsed, handlers), args, log_handler)
    return state


def make_config(parsed: Parsed, pos_args: Optional[Sequence[str]]) -> Config:
    """Make a tox configuration object."""
    # for now only tox.ini supported, assume empty tox.ini where pyproject.toml or in cwd
    tox_ini: Optional[Path] = parsed.config_file
    work_dir: Optional[Path] = parsed.work_dir
    if work_dir is not None:
        work_dir = work_dir.absolute()
    if tox_ini is None:
        folder = Path.cwd() if work_dir is None else work_dir
        candidate_names = "tox.ini", "pyproject.toml"
        for name in candidate_names:
            for base in chain([folder], folder.parents):
                candidate: Path = base / name
                if candidate.exists() and candidate.is_file():
                    tox_ini = candidate
                    break
            if tox_ini is not None:
                break
        if tox_ini is None:
            tox_ini = folder / "tox.ini"
            logging.warning(f"No {' or '.join(candidate_names)} found, assuming empty tox.ini at {tox_ini}")
        if work_dir is None:
            work_dir = tox_ini.parent
    else:
        tox_ini = tox_ini.absolute()
        work_dir = tox_ini.parent if work_dir is None else work_dir

    ini_loader = ToxIni(tox_ini)
    return Config(
        config_source=ini_loader,
        overrides=parsed.override,
        pos_args=pos_args,
        root=parsed.root_dir if parsed.root_dir is not None else tox_ini.parent,
        work_dir=work_dir,
    )



from tox.pytest import ToxProjectCreator


def test_depends(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.ini": """
    [tox]
    env_list = py39,py38,py37,cov2,cov
    [testenv]
    package = wheel
    [testenv:cov]
    depends = py39,py38,py37
    skip_install = true
    [testenv:cov2]
    depends = cov
    skip_install = true
    """
        }
    )
    outcome = project.run("de")
    outcome.assert_success()
    out = """Execution order: py39, py38, py37, cov, cov2
ALL
   py39 ~ .package-py39
   py38 ~ .package-py38
   py37 ~ .package-py37
   cov2
      cov
         py39 ~ .package-py39
         py38 ~ .package-py38
         py37 ~ .package-py37
   cov
      py39 ~ .package-py39
      py38 ~ .package-py38
      py37 ~ .package-py37
    """
    outcome.assert_out_err(out, "")



"""
A tox build environment that handles Python packages.
"""
import sys
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

from packaging.requirements import Requirement

from tox.config.main import Config

from ..package import PackageToxEnv
from .api import Python, PythonDep


class PythonPackage(Python, PackageToxEnv, ABC):
    def setup(self) -> None:
        """setup the tox environment"""
        super().setup()

        requires = [PythonDep(i) for i in self.requires()]
        self.cached_install(requires, PythonPackage.__name__, "requires")

        build_requires = [PythonDep(i) for i in self.build_requires()]
        self.cached_install(build_requires, PythonPackage.__name__, "build_requires")

    @abstractmethod
    def requires(self) -> Tuple[Requirement, ...]:
        raise NotImplementedError

    @abstractmethod
    def build_requires(self) -> Tuple[Requirement, ...]:
        raise NotImplementedError

    def default_base_python(self, conf: Config, env_name: Optional[str]) -> List[str]:
        return [sys.executable]



import subprocess
import sys
from pathlib import Path

from tox.pytest import ToxProject


def test_call_as_module(empty_project: ToxProject) -> None:
    subprocess.check_output([sys.executable, "-m", "tox", "-h"])


def test_call_as_exe(empty_project: ToxProject) -> None:
    subprocess.check_output([str(Path(sys.executable).parent / "tox4"), "-h"])



from collections import OrderedDict
from pathlib import Path
from typing import Callable, Dict, Optional, Set, TypeVar

import pytest

from tests.conftest import ToxIniCreator
from tox.config.sets import ConfigSet

ConfBuilder = Callable[[str], ConfigSet]


@pytest.fixture(name="conf_builder")
def _conf_builder(tox_ini_conf: ToxIniCreator) -> ConfBuilder:
    def _make(conf_str: str) -> ConfigSet:
        return tox_ini_conf(f"[tox]\nenvlist=py39\n[testenv]\n{conf_str}").get_env("py39")

    return _make


def test_config_str(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("deps-x = 1\n    other: 2")
    config_set.add_config(keys="deps-x", of_type=str, default="", desc="ok")
    result = config_set["deps-x"]
    assert result == "1"


def test_config_path(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("path = path")
    config_set.add_config(keys="path", of_type=Path, default=Path(), desc="path")
    path_materialize = config_set["path"]
    assert path_materialize == Path("path")


def test_config_set(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("set = 1\n 2\n 3")
    config_set.add_config(keys="set", of_type=Set[int], default=set(), desc="set")
    set_materialize = config_set["set"]
    assert set_materialize == {1, 2, 3}


def test_config_optional_none(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    config_set.add_config(
        keys="optional_none", of_type=Optional[int], default=None, desc="optional_none"  # type: ignore[arg-type]
    )
    optional_none = config_set["optional_none"]
    assert optional_none is None


def test_config_dict(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("dict = a=1\n  b=2\n  c=3")
    config_set.add_config(keys="dict", of_type=Dict[str, int], default=dict(), desc="dict")
    dict_val = config_set["dict"]
    assert dict_val == OrderedDict([("a", 1), ("b", 2), ("c", 3)])


def test_config_bad_type(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("crazy = something-bad")

    config_set.add_config(keys="crazy", of_type=TypeVar, default=TypeVar("V"), desc="crazy")
    with pytest.raises(TypeError) as context:
        assert config_set["crazy"]
    assert str(context.value) == f"something-bad cannot cast to {TypeVar!r}"


def test_config_bad_dict(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("bad_dict = something")

    config_set.add_config(keys="bad_dict", of_type=Dict[str, str], default={}, desc="bad_dict")
    with pytest.raises(TypeError) as context:
        assert config_set["bad_dict"]
    assert str(context.value) == "dictionary lines must be of form key=value, found something"


def test_config_bad_bool(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("bad_bool = whatever")
    config_set.add_config(keys="bad_bool", of_type=bool, default=False, desc="bad_bool")
    with pytest.raises(TypeError) as context:
        assert config_set["bad_bool"]
    error = "value whatever cannot be transformed to bool, valid: , 0, 1, false, no, off, on, true, yes"
    assert str(context.value) == error


def test_config_constant(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    config_set.add_constant(keys="a", value=1, desc="ok")
    const = config_set["a"]
    assert const == 1


def test_config_lazy_constant(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    config_set.add_constant(keys="b", value=lambda: 2, desc="ok")
    lazy_const = config_set["b"]
    assert lazy_const == 2


def test_config_constant_repr(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    defined = config_set.add_constant(keys="a", value=1, desc="ok")
    assert repr(defined)


def test_config_dynamic_repr(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("path = path")
    defined = config_set.add_config(keys="path", of_type=Path, default=Path(), desc="path")
    assert repr(defined)


def test_config_redefine_constant_fail(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("path = path")
    config_set.add_constant(keys="path", desc="desc", value="value")
    with pytest.raises(ValueError, match="config path already defined"):
        config_set.add_constant(keys="path", desc="desc", value="value")


def test_config_redefine_dynamic_fail(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("path = path")
    config_set.add_config(keys="path", of_type=str, default="default", desc="path")
    with pytest.raises(ValueError, match="config path already defined"):
        config_set.add_config(keys="path", of_type=str, default="default", desc="path")


def test_config_dynamic_not_equal(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    path = config_set.add_config(keys="path", of_type=Path, default=Path(), desc="path")
    paths = config_set.add_config(keys="paths", of_type=Path, default=Path(), desc="path")
    assert path != paths



from pathlib import Path
from typing import Tuple

from packaging.requirements import Requirement

from tox.plugin.impl import impl
from tox.tox_env.register import ToxEnvRegister

from ..api import Pep517VirtualEnvPackage


class Pep517VirtualEnvPackageSdist(Pep517VirtualEnvPackage):
    @staticmethod
    def id() -> str:
        return "virtualenv-pep-517-sdist"

    def build_requires(self) -> Tuple[Requirement, ...]:
        result = self.get_requires_for_build_sdist()
        return result.requires

    def _build_artifact(self) -> Path:
        result = self.build_sdist(sdist_directory=self.pkg_dir)
        return result.sdist


@impl
def tox_register_tox_env(register: ToxEnvRegister) -> None:
    register.add_package_env(Pep517VirtualEnvPackageSdist)



import pytest

from tests.conftest import ToxIniCreator
from tox.config.loader.api import Override
from tox.config.loader.memory import MemoryLoader
from tox.config.main import Config
from tox.config.sets import ConfigSet


@pytest.fixture
def empty_config(tox_ini_conf: ToxIniCreator) -> Config:
    return tox_ini_conf("")


def test_empty_config_repr(empty_config: Config) -> None:
    text = repr(empty_config)
    assert str(empty_config.core["tox_root"]) in text
    assert "config_source=ToxIni" in text


def test_empty_conf_tox_envs(empty_config: Config) -> None:
    tox_env_keys = list(empty_config)
    assert tox_env_keys == []


def test_empty_conf_get(empty_config: Config) -> None:
    result = empty_config.get_env("magic")
    assert isinstance(result, ConfigSet)
    loaders = result["base"]
    assert loaders == ["testenv"]


def test_config_some_envs(tox_ini_conf: ToxIniCreator) -> None:
    example = """
    [tox]
    env_list = py38, py37
    [testenv]
    deps = 1
        other: 2
    [testenv:magic]
    """
    config = tox_ini_conf(example)
    tox_env_keys = list(config)
    assert tox_env_keys == ["py38", "py37", "other", "magic"]

    config_set = config.get_env("py38")
    assert repr(config_set)
    assert isinstance(config_set, ConfigSet)
    assert list(config_set)


def test_config_overrides(tox_ini_conf: ToxIniCreator) -> None:
    conf = tox_ini_conf("[testenv]", override=[Override("testenv.c=ok")]).get_env("py")
    conf.add_config("c", of_type=str, default="d", desc="desc")
    assert conf["c"] == "ok"


def test_config_new_source(tox_ini_conf: ToxIniCreator) -> None:
    main_conf = tox_ini_conf("[testenv]", override=[Override("testenv.c=ok")])
    conf = main_conf.get_env("py", loaders=[MemoryLoader(c="something_else")])
    conf.add_config("c", of_type=str, default="d", desc="desc")
    assert conf["c"] == "something_else"



from tests.conftest import ToxIniCreator


def test_tox_ini_core(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf("[tox]")
    core_loader_1 = list(config._src.get_core({}))
    assert len(core_loader_1) == 1

    core_loader_2 = list(config._src.get_core({}))
    assert len(core_loader_2) == 1

    assert core_loader_1[0] is core_loader_2[0]



from argparse import ArgumentParser
from typing import Any, Iterator, List, Optional, Union

from tox.config.loader.str_convert import StrConvert


class CliEnv:
    def __init__(self, value: Union[None, List[str], str] = None):
        if isinstance(value, str):
            value = StrConvert().to(value, of_type=List[str])
        self.all: bool = value is None or "ALL" in value
        self._names = value

    def __iter__(self) -> Iterator[str]:
        if self._names is not None:  # pragma: no branch
            yield from self._names

    def __str__(self) -> str:
        return "ALL" if self.all or self._names is None else ",".join(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({'' if self.all else repr(self._names)})"

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.all == other.all and self._names == other._names

    def __ne__(self, other: Any) -> bool:
        return not (self == other)


def env_list_flag(parser: ArgumentParser, default: Optional[CliEnv] = None) -> None:
    parser.add_argument(
        "-e",
        dest="env",
        help="tox environment(s) to run (ALL -> all default environments)",
        default=CliEnv() if default is None else default,
        type=CliEnv,
    )



from tox.pytest import ToxProjectCreator


def test_quickstart(tox_project: ToxProjectCreator) -> None:
    project = tox_project({})
    outcome = project.run("q")
    outcome.assert_success()
    out = (
        f"ROOT: No tox.ini or pyproject.toml found, assuming empty tox.ini at {project.path / 'tox.ini'}\n"
        "done quickstart\n"
    )
    outcome.assert_out_err(out, "")



from threading import Lock


class AtomicCounter:
    def __init__(self) -> None:
        self.value: int = 0
        self._lock = Lock()

    def increment(self) -> None:
        with self._lock:
            self.value += 1

    def decrement(self) -> None:
        with self._lock:
            self.value -= 1



"""
A reader that drain a stream via its file no on a background thread.
"""
from abc import ABC, abstractmethod
from threading import Event, Thread
from types import TracebackType
from typing import Callable, Optional, Type

WAIT_GENERAL = 0.05  # stop thread join every so often (give chance to a signal interrupt)


class ReadViaThread(ABC):
    def __init__(self, file_no: int, handler: Callable[[bytes], None], name: str, drain: bool) -> None:
        self.file_no = file_no
        self.stop = Event()
        self.thread = Thread(target=self._read_stream, name=f"tox-r-{name}-{file_no}")
        self.handler = handler
        self._on_exit_drain = drain

    def __enter__(self) -> "ReadViaThread":
        self.thread.start()
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        self.stop.set()  # signal thread to stop
        while self.thread.is_alive():  # wait until it stops
            self.thread.join(WAIT_GENERAL)
        self._drain_stream()  # read anything left

    @abstractmethod
    def _read_stream(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def _drain_stream(self) -> None:
        raise NotImplementedError



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Generator, List, Optional, Tuple, cast

from tox.config.sets import ConfigSet
from tox.config.types import Command, EnvList
from tox.journal import EnvJournal
from tox.report import ToxHandler

from .api import ToxEnv
from .package import PackageToxEnv

if TYPE_CHECKING:
    from tox.config.cli.parser import Parsed


class RunToxEnv(ToxEnv, ABC):
    def __init__(
        self, conf: ConfigSet, core: ConfigSet, options: "Parsed", journal: EnvJournal, log_handler: ToxHandler
    ) -> None:
        self.has_package = False
        self.package_env: Optional[PackageToxEnv] = None
        super().__init__(conf, core, options, journal, log_handler)

    def register_config(self) -> None:
        super().register_config()
        self.conf.add_config(
            keys=["description"],
            of_type=str,
            default="",
            desc="description attached to the tox environment",
        )
        self.conf.add_config(
            keys=["commands"],
            of_type=List[Command],
            default=[],
            desc="the commands to be called for testing",
        )
        self.conf.add_config(
            keys=["commands_pre"],
            of_type=List[Command],
            default=[],
            desc="the commands to be called before testing",
        )
        self.conf.add_config(
            keys=["commands_post"],
            of_type=List[Command],
            default=[],
            desc="the commands to be called after testing",
        )
        self.conf.add_config(
            keys=["change_dir", "changedir"],
            of_type=Path,
            default=lambda conf, name: cast(Path, conf.core["tox_root"]),
            desc="Change to this working directory when executing the test command.",
        )
        self.conf.add_config(
            "depends",
            of_type=EnvList,
            desc="tox environments that this environment depends on (must be run after those)",
            default=EnvList([]),
        )
        self.has_package = self.add_package_conf()

    def add_package_conf(self) -> bool:
        """If this returns True package_env and package_tox_env_type configurations must be defined"""
        self.core.add_config(
            keys=["no_package", "skipsdist"],
            of_type=bool,
            default=False,
            desc="Is there any packaging involved in this project.",
        )
        core_no_package: bool = self.core["no_package"]
        if core_no_package is True:
            return False
        self.conf.add_config(
            keys="skip_install",
            of_type=bool,
            default=False,
            desc="skip installation",
        )
        skip_install: bool = self.conf["skip_install"]
        return not skip_install

    def set_package_env(self) -> Generator[Tuple[str, str], PackageToxEnv, None]:
        if not self.has_package:
            return
        of_type = self.conf["package_tox_env_type"]
        name = self.conf["package_env"]
        package_tox_env = yield name, of_type
        self.package_env = package_tox_env
        self.package_env.ref_count.increment()

    def clean(self, force: bool = False) -> None:
        super().clean(force)
        if self.package_env is not None:  # pragma: no cover branch
            with self.package_env.display_context(suspend=self.has_display_suspended):
                self.package_env.clean()  # do not pass force along, allow package env to ignore if requested

    @property
    def environment_variables(self) -> Dict[str, str]:
        environment_variables = super().environment_variables
        if self.has_package:  # if package(s) have been built insert them as environment variable
            if self.packages:
                environment_variables["TOX_PACKAGE"] = os.pathsep.join(self.packages)
        return environment_variables

    @property
    @abstractmethod
    def packages(self) -> List[str]:
        """:returns: a list of packages installed in the environment"""
        raise NotImplementedError

    def teardown(self) -> None:
        super().teardown()
        if self.package_env is not None:
            with self.package_env.display_context(suspend=self.has_display_suspended):
                self.package_env.teardown()

    def interrupt(self) -> None:
        super().interrupt()
        if self.package_env is not None:  # pragma: no branch
            self.package_env.interrupt()



"""Record information about tox environments"""
from typing import Any, Dict, List, Tuple

from tox.execute import Outcome


class EnvJournal:
    """Report the status of a tox environment"""

    def __init__(self, enabled: bool, name: str) -> None:
        self._enabled = enabled
        self.name = name
        self._content: Dict[str, Any] = {}
        self._executes: List[Tuple[str, Outcome]] = []

    def __setitem__(self, key: str, value: Any) -> None:
        self._content[key] = value

    @property
    def content(self) -> Dict[str, Any]:
        tests: List[Dict[str, Any]] = []
        setup: List[Dict[str, Any]] = []
        for run_id, outcome in self._executes:
            one = {
                "command": outcome.cmd,
                "output": outcome.out,
                "err": outcome.err,
                "retcode": outcome.exit_code,
                "elapsed": outcome.elapsed,
                "show_on_standard": outcome.show_on_standard,
                "run_id": run_id,
                "start": outcome.start,
                "end": outcome.end,
            }
            if run_id.startswith("commands") or run_id.startswith("build"):
                tests.append(one)
            else:
                setup.append(one)
        if tests:
            self["test"] = tests
        if setup:
            self["setup"] = setup
        return self._content

    def __bool__(self) -> bool:
        return self._enabled

    def add_execute(self, outcome: Outcome, run_id: str) -> None:
        self._executes.append((run_id, outcome))


__all__ = ("EnvJournal",)



"""Contains the plugin manager object"""
from typing import List, Type, cast

import pluggy

from tox import provision
from tox.config.cli.parser import ToxParser
from tox.config.loader import api as loader_api
from tox.config.sets import ConfigSet
from tox.session import state
from tox.session.cmd import depends, devenv, legacy, list_env, quickstart, show_config, version_flag
from tox.session.cmd.run import parallel, sequential
from tox.tox_env import package as package_api
from tox.tox_env.api import ToxEnv
from tox.tox_env.python.virtual_env import runner
from tox.tox_env.python.virtual_env.package.artifact import dev, sdist, wheel
from tox.tox_env.register import REGISTER, ToxEnvRegister

from . import NAME, spec


class Plugin:
    def __init__(self) -> None:
        self.manager: pluggy.PluginManager = pluggy.PluginManager(NAME)  # type: ignore[no-any-unimported]
        self.manager.add_hookspecs(spec)

        internal_plugins = (
            loader_api,
            provision,
            runner,
            dev,
            sdist,
            wheel,
            legacy,
            version_flag,
            quickstart,
            show_config,
            devenv,
            list_env,
            depends,
            parallel,
            sequential,
            package_api,
        )

        for plugin in internal_plugins:
            self.manager.register(plugin)
        self.manager.load_setuptools_entrypoints(NAME)
        self.manager.register(state)

        REGISTER.populate(self)
        self.manager.check_pending()

    def tox_add_option(self, parser: ToxParser) -> None:
        self.manager.hook.tox_add_option(parser=parser)

    def tox_add_core_config(self, core: ConfigSet) -> None:
        self.manager.hook.tox_add_core_config(core=core)

    def tox_register_tox_env(self, register: "ToxEnvRegister") -> List[Type[ToxEnv]]:
        return cast(List[Type[ToxEnv]], self.manager.hook.tox_register_tox_env(register=register))


MANAGER = Plugin()

__all__ = (
    "MANAGER",
    "Plugin",
)



import logging
import os
import signal
import sys
from io import TextIOWrapper
from pathlib import Path
from types import FrameType
from typing import Optional

from tox.execute import Outcome
from tox.execute.local_sub_process import LocalSubProcessExecutor
from tox.execute.request import ExecuteRequest, StdinSource
from tox.report import NamedBytesIO

logging.basicConfig(level=logging.DEBUG, format="%(relativeCreated)d\t%(levelname).1s\t%(message)s")
bad_process = Path(__file__).parent / "bad_process.py"

executor = LocalSubProcessExecutor(colored=False)
request = ExecuteRequest(
    cmd=[sys.executable, bad_process, sys.argv[1]],
    cwd=Path().absolute(),
    env=os.environ.copy(),
    stdin=StdinSource.API,
    run_id="",
)
out_err = TextIOWrapper(NamedBytesIO("out")), TextIOWrapper(NamedBytesIO("err"))


def show_outcome(outcome: Optional[Outcome]) -> None:
    if outcome is not None:  # pragma: no branch
        print(outcome.exit_code)
        print(repr(outcome.out))
        print(repr(outcome.err))
        print(outcome.elapsed, end="")
        print("done show outcome", file=sys.stderr)


def handler(s: signal.Signals, f: FrameType) -> None:
    logging.info(f"signal {s} at {f}")
    global interrupt_done
    if interrupt_done is False:  # pragma: no branch
        interrupt_done = True
        logging.info(f"interrupt via {status}")
        status.interrupt()
        logging.info(f"interrupt finished via {status}")


interrupt_done = False
signal.signal(signal.SIGINT, handler)
logging.info("PID %d start %r", os.getpid(), request)
try:
    with executor.call(request, show=False, out_err=out_err) as status:
        logging.info("wait on %r", status)
        while status.exit_code is None:
            status.wait()
        logging.info("wait over on %r", status)
    show_outcome(status.outcome)
except Exception as exception:  # pragma: no cover
    logging.exception(exception)  # pragma: no cover
finally:
    logging.info("done")
    logging.shutdown()



import pytest

from tox.config.cli.parse import get_options
from tox.pytest import CaptureFixture


def test_help_does_not_default_cmd(capsys: CaptureFixture) -> None:
    with pytest.raises(SystemExit):
        get_options("-h")
    out, err = capsys.readouterr()
    assert not err
    assert "--verbose" in out
    assert "command:" in out



import os
import sys
from contextlib import contextmanager
from pathlib import Path
from subprocess import PIPE, Popen
from threading import Thread
from typing import IO, Any, Iterator, Optional, Tuple, cast

from packaging.requirements import Requirement

from .frontend import CmdStatus, Frontend


class SubprocessCmdStatus(CmdStatus, Thread):
    def __init__(self, process: "Popen[str]") -> None:
        super().__init__()
        self.process = process
        self._out_err: Optional[Tuple[str, str]] = None
        self.start()

    def run(self) -> None:
        self._out_err = self.process.communicate()

    @property
    def done(self) -> bool:
        return self.process.returncode is not None

    def out_err(self) -> Tuple[str, str]:
        return cast(Tuple[str, str], self._out_err)


class SubprocessFrontend(Frontend):
    def __init__(
        self,
        root: Path,
        backend_paths: Tuple[Path, ...],
        backend_module: str,
        backend_obj: Optional[str],
        requires: Tuple[Requirement, ...],
    ):
        super().__init__(root, backend_paths, backend_module, backend_obj, requires, reuse_backend=False)

    @contextmanager
    def _send_msg(  # type: ignore[override]
        self, cmd: str, result_file: Path, msg: str
    ) -> Iterator[SubprocessCmdStatus]:
        env = os.environ.copy()
        env["PYTHONPATH"] = os.pathsep.join(str(i) for i in self._backend_paths)
        process = Popen(
            args=[sys.executable] + self.backend_args,
            stdout=PIPE,
            stderr=PIPE,
            stdin=PIPE,
            universal_newlines=True,
            cwd=self._root,
            env=env,
        )
        cast(IO[str], process.stdin).write(f"{os.linesep}{msg}{os.linesep}")
        yield SubprocessCmdStatus(process)

    def send_cmd(self, cmd: str, **kwargs: Any) -> Tuple[Any, str, str]:
        return self._send(cmd, **kwargs)


__all__ = (
    "SubprocessCmdStatus",
    "SubprocessFrontend",
)



from typing import Dict, List

from tox.config.cli.parser import ToxParser
from tox.plugin.impl import impl
from tox.session.cmd.run.common import run_order
from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    parser.add_command(
        "depends",
        ["de"],
        "visualize tox environment dependencies",
        depends,
    )


def depends(state: State) -> int:
    to_run_list = list(state.env_list())
    order, todo = run_order(state, to_run_list)
    print(f"Execution order: {', '.join(order)}")

    deps: Dict[str, List[str]] = {k: [o for o in order if o in v] for k, v in todo.items()}
    deps["ALL"] = to_run_list

    def _handle(at: int, env: str) -> None:
        print("   " * at, end="")
        print(env, end="")
        if env != "ALL":
            pkg_env = state.tox_env(env).package_env
            if pkg_env is not None:
                print(f" ~ {pkg_env.conf['env_name']}", end="")
        print("")
        at += 1
        for dep in deps[env]:
            _handle(at, dep)

    _handle(0, "ALL")
    return 0



"""
A tox python environment runner that uses the virtualenv project.
"""
from pathlib import Path
from typing import Dict, Optional, Set

from tox.config.main import Config
from tox.plugin.impl import impl
from tox.report import HandledError
from tox.tox_env.register import ToxEnvRegister

from ..runner import PythonRun
from .api import VirtualEnv
from .package.api import PackageType
from .package.util import virtual_env_package_id


class VirtualEnvRunner(VirtualEnv, PythonRun):
    """local file system python virtual environment via the virtualenv package"""

    @staticmethod
    def id() -> str:
        return "virtualenv"

    def add_package_conf(self) -> bool:
        if not super().add_package_conf():
            return False
        self.conf.add_config(keys="usedevelop", desc="use develop mode", default=False, of_type=bool)
        develop_mode = self.conf["usedevelop"]
        desc = f"package installation mode - {' | '.join(i.name for i in PackageType)} "
        if develop_mode:
            self.conf.add_constant(["package"], desc, PackageType.dev)
        else:
            self.conf.add_config(keys="package", of_type=PackageType, default=PackageType.sdist, desc=desc)
        try:
            pkg_type: PackageType = self.conf["package"]
        except AttributeError as exc:
            values = ", ".join(i.name for i in PackageType)
            error = HandledError(f"invalid package config type {exc.args[0]!r} requested, must be one of {values}")
            raise error from exc

        if pkg_type == PackageType.skip:
            return False

        self.conf.add_constant(
            keys=["package_tox_env_type"],
            desc="tox package type used to package",
            value=virtual_env_package_id(pkg_type),
        )

        def default_package_name(conf: Config, name: Optional[str]) -> str:
            result = ".package"

            # when building wheels we need to ensure that the built package is compatible with the target env
            # compatibility is documented within https://www.python.org/dev/peps/pep-0427/#file-name-convention
            # a wheel tag example: {distribution}-{version}(-{build tag})?-{python tag}-{abi tag}-{platform tag}.whl
            # python only code are often compatible at major level (unless universal wheel in which case both 2/3)
            # c-extension codes are trickier, but as of today both poetry/setuptools uses pypa/wheels logic
            # https://github.com/pypa/wheel/blob/master/src/wheel/bdist_wheel.py#L234-L280
            # technically, the build tags can be passed in as CLI args to the build backend, but for now it's easier
            # to just create a new build env for every target env
            if pkg_type is PackageType.wheel:
                result = f"{result}-{self.conf['env_name']}"
            return result

        self.conf.add_config(
            keys=["package_env", "isolated_build_env"],
            of_type=str,
            default=default_package_name,
            desc="tox environment used to package",
        )
        self.conf.add_config(
            keys=["extras"],
            of_type=Set[str],
            default=set(),
            desc="extras to install of the target package",
        )
        return True

    def install_package_args(self) -> Dict[str, bool]:
        return {
            "no_deps": True,  # dependencies are installed separately
            "develop": self.conf["package"] is PackageType.dev,  # if package type is develop mode pass option through
            "force_reinstall": True,  # if is already installed reinstall
        }

    def install_requirement_file(self, path: Path) -> None:
        install_command = self.base_install_cmd + ["-r", str(path)]
        result = self.perform_install(install_command, "install_deps")
        result.assert_success()


@impl
def tox_register_tox_env(register: ToxEnvRegister) -> None:
    register.add_run_env(VirtualEnvRunner)



"""
A tox run environment that handles the Python language.
"""
import logging
from abc import ABC, abstractmethod
from hashlib import sha256
from pathlib import Path
from typing import Any, Dict, List, Set

from tox.config.cli.parser import Parsed
from tox.config.sets import ConfigSet
from tox.journal import EnvJournal
from tox.report import ToxHandler
from tox.tox_env.errors import Recreate

from ..runner import RunToxEnv
from .api import Python, PythonDep
from .req_file import RequirementsFile


class PythonRun(Python, RunToxEnv, ABC):
    def __init__(self, conf: ConfigSet, core: ConfigSet, options: Parsed, journal: EnvJournal, log_handler: ToxHandler):
        super().__init__(conf, core, options, journal, log_handler)
        self._packages: List[PythonDep] = []

    def register_config(self) -> None:
        super().register_config()
        self.conf.add_config(
            keys="deps",
            of_type=RequirementsFile,
            default=RequirementsFile(""),
            desc="Name of the python dependencies as specified by PEP-440",
        )
        self.core.add_config(
            keys=["skip_missing_interpreters"],
            default=True,
            of_type=bool,
            desc="skip running missing interpreters",
        )

    def setup(self) -> None:
        """setup the tox environment"""
        super().setup()
        self.install_deps()

        if self.package_env is None:
            return
        skip_pkg_install: bool = getattr(self.options, "skip_pkg_install", False)
        if skip_pkg_install is True:
            logging.warning("skip building and installing the package")
            return

        # 1. install pkg dependencies
        with self.package_env.display_context(suspend=self.has_display_suspended):
            package_deps = self.package_env.get_package_dependencies(self.conf["extras"])
        self.cached_install([PythonDep(p) for p in package_deps], PythonRun.__name__, "package_deps")

        # 2. install the package
        with self.package_env.display_context(suspend=self.has_display_suspended):
            self._packages = [PythonDep(p) for p in self.package_env.perform_packaging()]
        self.install_python_packages(
            self._packages, "package", **self.install_package_args()  # type: ignore[no-untyped-call]
        )
        self.handle_journal_package(self.journal, self._packages)

    def install_deps(self) -> None:
        requirements_file: RequirementsFile = self.conf["deps"]
        requirement_file_content = requirements_file.validate_and_expand()
        requirement_file_content.sort()  # stable order dependencies
        with self._cache.compare(requirement_file_content, PythonRun.__name__, "deps") as (eq, old):
            if not eq:
                # if new env, or additions only a simple install will do
                missing: Set[str] = set() if old is None else set(old) - set(requirement_file_content)
                if not missing:
                    if requirement_file_content:
                        with requirements_file.with_file() as path:
                            self.install_requirement_file(path)
                else:  # otherwise, no idea how to compute the diff, instead just start from scratch
                    logging.warning(f"recreate env because dependencies removed: {', '.join(str(i) for i in missing)}")
                    raise Recreate

    @abstractmethod
    def install_package_args(self) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def install_requirement_file(self, path: Path) -> None:
        raise NotImplementedError

    @property
    def packages(self) -> List[str]:
        return [str(d.value) for d in self._packages]

    @staticmethod
    def handle_journal_package(journal: EnvJournal, package: List[PythonDep]) -> None:
        if not journal:
            return
        installed_meta = []
        for dep in package:
            if isinstance(dep.value, Path):
                pkg = dep.value
                of_type = "file" if pkg.is_file() else ("dir" if pkg.is_dir() else "N/A")
                meta = {"basename": pkg.name, "type": of_type}
                if of_type == "file":
                    meta["sha256"] = sha256(pkg.read_bytes()).hexdigest()
                installed_meta.append(meta)
        if installed_meta:
            journal["installpkg"] = installed_meta[0] if len(installed_meta) == 1 else installed_meta



import logging
import os
import re
from contextlib import contextmanager
from pathlib import Path
from tempfile import mkstemp
from typing import Iterator, List, Optional

from packaging.requirements import InvalidRequirement, Requirement

LOGGER = logging.getLogger(__name__)


VCS = ["ftp", "ssh", "git", "hg", "bzr", "sftp", "svn"]
VALID_SCHEMAS = ["http", "https", "file"] + VCS


def is_url(name: str) -> bool:
    return get_url_scheme(name) in VALID_SCHEMAS


def get_url_scheme(url: str) -> Optional[str]:
    return None if ":" not in url else url.split(":", 1)[0].lower()


class RequirementsFile:
    """
    Specification is defined within pip itself and documented under:
    - https://pip.pypa.io/en/stable/reference/pip_install/#requirements-file-format
    - https://github.com/pypa/pip/blob/master/src/pip/_internal/req/constructors.py#L291
    """

    VALID_OPTIONS = {
        "no_arg": [
            "--no-index",
            "--prefer-binary",
            "--require-hashes",
            "--pre",
        ],
        "one_arg": [
            "-i",
            "--index-url",
            "--extra-index-url",
            "-e",
            "--editable",
            "-c",
            "--constraint",
            "-r",
            "--requirement",
            "-f",
            "--find-links",
            "--trusted-host",
            "--use-feature",
            "--no-binary",
            "--only-binary",
        ],
    }

    def __init__(self, raw: str, allow_short_req_file: bool = True, root: Optional[Path] = None) -> None:
        self._root = Path().cwd() if root is None else root
        if allow_short_req_file:  # patch tox supporting requirements files via -rrequirements.txt
            r = ((f"-r {i[2:]}" if len(i) >= 3 and i.startswith("-r") and i[2] != " " else i) for i in raw.splitlines())
            adjusted = "\n".join(r)
            raw = f"{adjusted}\n" if raw.endswith("\\\n") else adjusted
        self._raw = raw

    def __str__(self) -> str:
        return self._raw

    @property
    def root(self) -> Path:
        return self._root

    def validate_and_expand(self) -> List[str]:
        raw = self._normalize_raw()
        result: List[str] = []
        ini_dir = self.root
        for at, line in enumerate(raw.splitlines(), start=1):
            if line.startswith("#"):
                continue
            line = re.sub(r"\s#.*", "", line).strip()
            if not line:
                continue
            if line.startswith("-"):  # handle flags
                words = re.findall(r"\S+", line)
                first = words[0]
                if first in self.VALID_OPTIONS["no_arg"]:
                    if len(words) != 1:
                        raise ValueError(line)
                    else:
                        result.append(" ".join(words))
                elif first in self.VALID_OPTIONS["one_arg"]:
                    if len(words) != 2:
                        raise ValueError(line)
                    else:
                        if first == "-r":
                            path = Path(line[3:].strip())
                            if not path.is_absolute():
                                path = ini_dir / path
                            req_file = RequirementsFile(path.read_text(), allow_short_req_file=False, root=self.root)
                            result.extend(req_file.validate_and_expand())
                        else:
                            result.append(" ".join(words))
                else:
                    raise ValueError(line)
            else:
                try:
                    req = Requirement(line)
                    result.append(str(req))
                except InvalidRequirement as exc:
                    if is_url(line) or any(line.startswith(f"{v}+") and is_url(line[len(v) + 1 :]) for v in VCS):
                        result.append(line)
                    else:
                        path = ini_dir / line
                        try:
                            is_valid_file = path.exists() and path.is_file()
                        except OSError:  # https://bugs.python.org/issue42855 # pragma: no cover
                            is_valid_file = False  # pragma: no cover
                        if not is_valid_file:
                            raise ValueError(f"{at}: {line}") from exc
                        result.append(str(path))
        return result

    def _normalize_raw(self) -> str:
        # a line ending in an unescaped \ is treated as a line continuation and the newline following it is effectively
        # ignored
        raw = "".join(self._raw.replace("\r", "").split("\\\n"))
        # Since version 10, pip supports the use of environment variables inside the requirements file.
        # You can now store sensitive data (tokens, keys, etc.) in environment variables and only specify the variable
        # name for your requirements, letting pip lookup the value at runtime.
        # You have to use the POSIX format for variable names including brackets around the uppercase name as shown in
        # this example: ${API_TOKEN}. pip will attempt to find the corresponding environment variable defined on the
        # host system at runtime.
        while True:
            match = re.search(r"\$\{([A-Z_]+)\}", raw)
            if match is None:
                break
            value = os.environ.get(match.groups()[0], "")
            start, end = match.span()
            raw = f"{raw[:start]}{value}{raw[end:]}"
        return raw

    @contextmanager
    def with_file(self) -> Iterator[Path]:
        file_no, path = mkstemp(dir=self.root, prefix="requirements-", suffix=".txt")
        try:
            try:
                with open(path, "wt") as f:
                    f.write(self._raw)
            finally:
                os.close(file_no)
            yield Path(path)
        finally:
            os.unlink(path)



"""
Expand tox factor expressions to tox environment list.
"""
import itertools
import re
from typing import Iterator, List, Optional, Tuple


def filter_for_env(value: str, name: Optional[str]) -> str:
    current = (
        set(itertools.chain.from_iterable([(i for i, _ in a) for a in find_factor_groups(name)]))
        if name is not None
        else set()
    )
    overall = []
    for factors, content in expand_factors(value):
        if factors is None:
            overall.append(content)
        else:
            for group in factors:
                for a_name, negate in group:
                    contains = a_name in current
                    if contains == negate:
                        break
                else:
                    overall.append(content)
    result = "\n".join(overall)
    return result


def find_envs(value: str) -> Iterator[str]:
    seen = set()
    for factors, _ in expand_factors(value):
        if factors is not None:
            for group in factors:
                env = explode_factor(group)
                if env not in seen:
                    yield env
                    seen.add(env)


def extend_factors(value: str) -> Iterator[str]:
    for group in find_factor_groups(value):
        yield explode_factor(group)


def explode_factor(group: List[Tuple[str, bool]]) -> str:
    return "-".join([name for name, _ in group])


def expand_factors(value: str) -> Iterator[Tuple[Optional[Iterator[List[Tuple[str, bool]]]], str]]:
    for line in value.split("\n"):
        match = re.match(r"^((?P<factor_expr>[\w{}.!,-]+):\s+)?(?P<content>.*?)$", line)
        if match is None:  # pragma: no cover
            raise RuntimeError("for a valid factor regex this cannot happen")
        groups = match.groupdict()
        factor_expr, content = groups["factor_expr"], groups["content"]
        if factor_expr is not None:
            factors = find_factor_groups(factor_expr)
            yield factors, content
        else:
            yield None, content


def find_factor_groups(value: str) -> Iterator[List[Tuple[str, bool]]]:
    """transform '{py,!pi}-{a,b},c' to [{'py', 'a'}, {'py', 'b'}, {'pi', 'a'}, {'pi', 'b'}, {'c'}]"""
    for env in expand_env_with_negation(value):
        result = [name_with_negate(f) for f in env.split("-")]
        yield result


def expand_env_with_negation(value: str) -> Iterator[str]:
    """transform '{py,!pi}-{a,b},c' to ['py-a', 'py-b', '!pi-a', '!pi-b', 'c']"""
    for key, group in itertools.groupby(re.split(r"((?:{[^}]+})+)|,", value), key=bool):
        if key:
            group_str = "".join(group).strip()
            elements = re.split(r"{([^}]+)}", group_str)
            parts = [re.sub(r"\s+", "", elem).split(",") for elem in elements]
            for variant in itertools.product(*parts):
                variant_str = "".join(variant)
                yield variant_str


def name_with_negate(factor: str) -> Tuple[str, bool]:
    negated = is_negated(factor)
    result = factor[1:] if negated else factor
    return result, negated


def is_negated(factor: str) -> bool:
    return factor.startswith("!")


__all__ = (
    "filter_for_env",
    "find_envs",
    "expand_factors",
    "extend_factors",
)



import pytest

from tox.session.common import CliEnv


@pytest.mark.parametrize(
    "val, exp",
    [
        (CliEnv(["a", "b"]), "CliEnv(['a', 'b'])"),
        (CliEnv(["ALL", "b"]), "CliEnv()"),
        (CliEnv([]), "CliEnv([])"),
    ],
)
def test_cli_env_repr(val: CliEnv, exp: str) -> None:
    assert repr(val) == exp


def test_cli_env_repr_all() -> None:
    env = CliEnv(["ALL", "b"])
    assert repr(env) == "CliEnv()"


@pytest.mark.parametrize(
    "val, exp",
    [
        (CliEnv(["a", "b"]), "a,b"),
        (CliEnv(["ALL", "b"]), "ALL"),
        (CliEnv([]), ""),
    ],
)
def test_cli_env_str(val: CliEnv, exp: str) -> None:
    assert str(val) == exp



import sys
import tarfile
from pathlib import Path
from textwrap import dedent
from zipfile import ZipFile

name = "demo_pkg_inline"
pkg_name = name.replace("_", "-")

version = "1.0.0"
dist_info = f"{name}-{version}.dist-info"

content = {
    f"{name}/__init__.py": f"def do():\n    print('greetings from {name}')",
    f"{dist_info}/METADATA": f"""
        Metadata-Version: 2.1
        Name: {pkg_name}
        Version: {version}
        Summary: UNKNOWN
        Home-page: UNKNOWN
        Author: UNKNOWN
        Author-email: UNKNOWN
        License: UNKNOWN
        Platform: UNKNOWN

        UNKNOWN
       """,
    f"{dist_info}/WHEEL": f"""
        Wheel-Version: 1.0
        Generator: {name}-{version}
        Root-Is-Purelib: true
        Tag: py3-none-any
       """,
    f"{dist_info}/top_level.txt": name,
    f"{dist_info}/RECORD": f"""
        {name}/__init__.py,,
        {dist_info}/METADATA,,
        {dist_info}/WHEEL,,
        {dist_info}/top_level.txt,,
        {dist_info}/RECORD,,
       """,
}


def build_wheel(wheel_directory, metadata_directory=None, config_settings=None):
    base_name = f"{name}-{version}-py{sys.version_info.major}-none-any.whl"
    path = Path(wheel_directory) / base_name
    with ZipFile(str(path), "w") as zip_file_handler:
        for arc_name, data in content.items():
            zip_file_handler.writestr(zinfo_or_arcname=arc_name, data=dedent(data).strip())
    print(f"created wheel {path}")
    return base_name


def get_requires_for_build_wheel(config_settings):
    return []  # pragma: no cover # only executed in non-host pythons


def build_sdist(sdist_directory, config_settings=None):
    result = f"{name}-{version}.tar.gz"
    with tarfile.open(str(Path(sdist_directory) / result), "w:gz") as tar:
        root = Path(__file__).parent
        tar.add(str(root / "build.py"), arcname="build.py")
        tar.add(str(root / "pyproject.toml"), arcname="pyproject.toml")
    return result


def get_requires_for_build_sdist(config_settings):
    return []  # pragma: no cover # only executed in non-host pythons



"""
Provides configuration values from the the environment variables.
"""
import logging
import os
from typing import Any, Optional, Tuple, Type

from tox.config.loader.str_convert import StrConvert

CONVERT = StrConvert()


def get_env_var(key: str, of_type: Type[Any]) -> Optional[Tuple[Any, str]]:
    """Get the environment variable option.

    :param key: the config key requested
    :param of_type: the type we would like to convert it to
    :return:
    """
    key_upper = key.upper()
    for fmt in ("TOX_{}", "TOX{}"):
        environ_key = fmt.format(key_upper)
        if environ_key in os.environ:
            value = os.environ[environ_key]
            try:
                source = f"env var {environ_key}"
                result = CONVERT.to(raw=value, of_type=of_type)
                return result, source
            except Exception as exception:  # noqa
                logging.warning(
                    "env var %s=%r cannot be transformed to %r because %r",
                    environ_key,
                    value,
                    of_type,
                    exception,
                )
    return None


__all__ = ("get_env_var",)



import json
import os
import sys
from pathlib import Path
from subprocess import check_call
from typing import List, Optional
from zipfile import ZipFile

import pytest
from packaging.requirements import Requirement

from tox.pytest import Index, IndexServer, MonkeyPatch, TempPathFactory, ToxProjectCreator

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import Distribution  # type: ignore[attr-defined]
else:  # pragma: no cover (<py38)
    from importlib_metadata import Distribution  # noqa

ROOT = Path(__file__).parents[1]


@pytest.fixture(scope="session")
def tox_wheel(tmp_path_factory: TempPathFactory) -> Path:
    # takes around 3.2s
    package: Optional[Path] = None
    if "TOX_PACKAGE" in os.environ:
        env_tox_pkg = Path(os.environ["TOX_PACKAGE"])
        if env_tox_pkg.exists() and env_tox_pkg.suffix == ".whl":
            package = env_tox_pkg
    if package is None:  # pragma: no cover
        # when we don't get a wheel path injected, build it (for example when running from an IDE)
        package = build_wheel(tmp_path_factory.mktemp("dist"), Path(__file__).parents[1])
    return package


@pytest.fixture(scope="session")
def tox_wheels(tox_wheel: Path, tmp_path_factory: TempPathFactory) -> List[Path]:
    # takes around 1.5s if already cached
    result: List[Path] = [tox_wheel]
    info = tmp_path_factory.mktemp("info")
    with ZipFile(str(tox_wheel), "r") as zip_file:
        zip_file.extractall(path=info)
    dist_info = next((i for i in info.iterdir() if i.suffix == ".dist-info"), None)
    if dist_info is None:  # pragma: no cover
        raise RuntimeError(f"no tox.dist-info inside {tox_wheel}")
    distribution = Distribution.at(dist_info)
    wheel_cache = ROOT / ".wheel_cache" / f"{sys.version_info.major}.{sys.version_info.minor}"
    wheel_cache.mkdir(parents=True, exist_ok=True)
    cmd = [sys.executable, "-I", "-m", "pip", "download", "-d", str(wheel_cache)]
    for req in distribution.requires:
        requirement = Requirement(req)
        if not requirement.extras:  # pragma: no branch  # we don't need to install any extras (tests/docs/etc)
            cmd.append(req)
    check_call(cmd)
    result.extend(wheel_cache.iterdir())
    return result


@pytest.fixture(scope="session")
def demo_pkg_inline_wheel(tmp_path_factory: TempPathFactory, demo_pkg_inline: Path) -> Path:
    return build_wheel(tmp_path_factory.mktemp("dist"), demo_pkg_inline)


def build_wheel(dist_dir: Path, of: Path) -> Path:
    from build.__main__ import build_package  # noqa

    build_package(str(of), str(dist_dir), distributions=["wheel"])
    package = next(dist_dir.iterdir())
    return package


@pytest.fixture(scope="session")
def pypi_index_self(pypi_server: IndexServer, tox_wheels: List[Path], demo_pkg_inline_wheel: Path) -> Index:
    # takes around 1s
    self_index = pypi_server.create_index("self", "volatile=False")
    self_index.upload(tox_wheels + [demo_pkg_inline_wheel])
    return self_index


def test_provision_requires_nok(tox_project: ToxProjectCreator) -> None:
    ini = "[tox]\nrequires = pkg-does-not-exist\n setuptools==1\nskipsdist=true\n"
    outcome = tox_project({"tox.ini": ini}).run("c", "-e", "py")
    outcome.assert_failed()
    outcome.assert_out_err(
        r".*will run in automatically provisioned tox, host .* is missing \[requires \(has\)\]:"
        r" pkg-does-not-exist \(N/A\), setuptools==1 \(.*\).*",
        r".*",
        regex=True,
    )


@pytest.mark.integration
@pytest.mark.timeout(60)
def test_provision_requires_ok(
    tox_project: ToxProjectCreator, pypi_index_self: Index, monkeypatch: MonkeyPatch, tmp_path: Path
) -> None:
    log = tmp_path / "out.log"
    pypi_index_self.use(monkeypatch)
    ini = "[tox]\nrequires = demo-pkg-inline\n setuptools \n[testenv]\npackage=skip"

    outcome = tox_project({"tox.ini": ini}).run("r", "-e", "py", "--result-json", str(log))

    outcome.assert_success()
    with log.open("rt") as file_handler:
        log_report = json.load(file_handler)
    assert "py" in log_report["testenvs"]
